/*
 *
 * FreeBSD install - a package for the installation and maintainance
 * of non-core utilities.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 * Jordan K. Hubbard
 * 18 July 1993
 *
 * This is the info module.
 *
 */

#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include "list.h"
#include "lib.h"

#ifndef lint
static const char rcsid[] =
  "$FreeBSD: src/usr.sbin/pkg_install/info/main.c,v 1.31 2001/03/23 18:45:24 sobomax Exp $";
#endif

static char Options[] = "acdDe:fghiIkl:LmO:opqrRsS:t:vW:";

static void usage __P((void));
void cleanup(int);

int
main(int argc, char **argv)
{
	int ch, err = 0;
	unsigned int flags = PKG_SHOW_NONE;
	boolean all = FALSE;
	which_list wl = { PLIST_HEAD_INITIALIZER };
	which_entry *w;
	pkg_entry *pe;

	signal(SIGINT, cleanup);
	signal(SIGHUP, cleanup);

	pkg_mgr_init();
	PkgMgr.verbose = TRUE;
	if (argc == 1) {
		all = TRUE;
		flags = PKG_SHOW_INDEX;
	}
	else while ((ch = getopt(argc, argv, Options)) != -1) {
		switch(ch) {
		case 'a':
			all = TRUE;
			break;

		case 'v':
			/* Reasonable definition of 'everything' */
			flags = PKG_SHOW_COMMENT | PKG_SHOW_DESC | PKG_SHOW_PLIST | PKG_SHOW_INSTALL |
				PKG_SHOW_DEINSTALL | PKG_SHOW_REQUIRE | PKG_SHOW_DISPLAY | PKG_SHOW_MTREE;
			break;

		case 'I':
			flags |= PKG_SHOW_INDEX;
			break;

		case 'p':
			flags |= PKG_SHOW_PREFIX;
			break;

		case 'c':
			flags |= PKG_SHOW_COMMENT;
			break;

		case 'd':
			flags |= PKG_SHOW_DESC;
			break;

		case 'D':
			flags |= PKG_SHOW_DISPLAY;
			break;

		case 'f':
			flags |= PKG_SHOW_PLIST;
			break;

		case 'g':
			flags |= PKG_SHOW_CHKSUM;
			break;

		case 'i':
			flags |= PKG_SHOW_INSTALL;
			break;

		case 'k':
			flags |= PKG_SHOW_DEINSTALL;
			break;

		case 'r':
			flags |= PKG_SHOW_REQUIRE;
			break;

		case 'R':
			flags |= PKG_SHOW_REQBY;
			break;

		case 'L':
			flags |= PKG_SHOW_FILES;
			break;

		case 'm':
			flags |= PKG_SHOW_MTREE;
			break;

		case 's':
			flags |= PKG_SHOW_SIZE;
			break;

		case 'o':
			flags |= PKG_SHOW_ORIGIN;
			break;

		case 'l':
			PkgMgr.infoprefix = optarg;
			break;

		case 'q':
			PkgMgr.verbose = FALSE;
			break;

		case 't':
			strlcpy(PkgMgr.mktmp_template, optarg, PATH_MAX);
			break;

		case 'e':
			if (pkg_mgr_check(optarg))
				return 0;
			else
				return 1;
			break;

		case 'W':
			flags |= PKG_SHOW_WHICH;
			if ((strrchr(optarg, '/') != NULL)
			 || (isnode(optarg) && isfile(optarg)))
				w = add_which_entry(&wl, WHICH_FILE, optarg);
			else
				w = add_which_entry(&wl, WHICH_EXEC, optarg);
			if (!w)
				msg_warn("%s not found", optarg);
			break;

		case 'S':
			flags |= PKG_SHOW_WHICH;
			w = add_which_entry(&wl, WHICH_LIBRARY, optarg);
			if (!w)
				msg_warn("%s not found", optarg);
			break;

		case 'O':
			flags |= PKG_SHOW_WHICH;
			if (add_which_plist(&wl, optarg)) {
				msg_error(2, "Unable to parse PLIST.");
				exit(2);
			}
			if (PLIST_EMPTY(&wl))
				msg_verbose("no files from plist found");
			break;

		case 'h':
		case '?':
		default:
			usage();
			break;
		}
	}
	argc -= optind;
	argv += optind;

	/*
	 * If we are only looking for which entries, and the list
	 * is empty, then we exit with an error.
     */	
	if (flags == PKG_SHOW_WHICH && PLIST_EMPTY(&wl))
		exit(1);
	scan_which(&wl);

	/* Set some reasonable defaults */
	if (!flags)
		flags = PKG_SHOW_COMMENT | PKG_SHOW_DESC | PKG_SHOW_REQBY;

	/* If no packages, yelp */
	if (!argc && !all) {
		if (flags & PKG_SHOW_WHICH)
			all = TRUE;
		else
			msg_warn("missing package name(s)"), usage();
	}	
	/* Get all the remaining package names, if any */
	if (all) {
		PLIST_FOREACH(pe, &PkgMgr)
			err += pkg_mgr_show(pe->dbdir, flags, &wl);
	} else {
		while (*argv) {
			err += pkg_mgr_show(*argv, flags, &wl);
			argv++;
		}
	}
	if (err != 0)
		msg_warn("%d package(s) failed", err);

	free_which(&wl);
	return err;
}

static void
usage()
{
	fprintf(stderr, "%s\n%s\n%s\n%s\n",
		"usage: pkg_info [-cdDfiIkLmoOpqrRsSvW] [-e package] [-l prefix]",
		"				[-t template] [pkg-name ...]",
		"	   pkg_info [-W file] [-S shlib] [-O]",
		"	   pkg_info -a [flags]");
	exit(1);
}

void
cleanup(int sig)
{
	static int in_cleanup = 0;

	if (!in_cleanup) {
		in_cleanup = 1;
		printf("Signal %d received, cleaning up..\n", sig);
	}
	if (sig)
		exit(1);
}
