/*
 * XXX: (reg) insert final project name here.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Jeremy D. Lea
 * XXX: (reg) Insert date
 *
 * Based on pkg_install, by:
 *   Jordan K. Hubbard
 *   18 July 1993
 *
 * This file contains code for managing lists of dependencies.
 */

#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "list.h"
#include <err.h>
#include "lib.h"

/*
 * Allocate a new depend list entry.
 */
depend_entry *
new_depend_entry(void)
{
	depend_entry *d;

	if ((d = (depend_entry *)malloc(sizeof(depend_entry))) == NULL) {
		msg_error(2, "%s: unable to allocate memory", __func__);
		return NULL;
	}	
	bzero(d, sizeof(depend_entry));
	return d;
}

/*
 * Free an entire depend list, and the piggy backed which list.
 */
void
free_depend(depend_list *dl)
{
	depend_entry *d;

	if (dl == NULL)
		return;
	free_which(&dl->wl);
	while (!PLIST_EMPTY(dl)) {
		PLIST_REMOVE_HEAD(dl, d);
		free(d);
	}
}

/*
 * Add an item to a depend list.
 */
depend_entry *
add_depend_entry(depend_list *dl, depend_t type, char *file, char *pkg)
{
	depend_entry *d;
	
	if (dl == NULL)
		return NULL;
	if ((d = new_depend_entry()) == NULL)
		return NULL;
	d->type = type;
	if (file != NULL && strlen(file) > 0)
		strlcpy(d->file, file, PATH_MAX);
	else {
		d->file[0] = '\0';
		if (type != DEPEND_PKG) {
			msg_warn("to depend on a file, we must have a file!");
			return NULL;
		}
	}	
	if (pkg != NULL && strlen(pkg) > 0)
		strlcpy(d->pkg, pkg, PATH_MAX);
	else {
		d->pkg[0] = '\0';
		if (type == DEPEND_PKG) {
			msg_warn("to depend on a package, we must have a package!");
			return NULL;
		}
	}
	/*
	 * Now that we know the file and/or package, we make a which_entry
	 * to check later if where the file really came from, if it exists.
	 * If we only have a package, then we're stuck with depending on
	 * the package explictly.
	 *
	 * We maintain a which_list piggy backed onto our depend_list, and
	 * pointers for each entry so that we can find corresponding entries.
	 */
	switch (d->type) {
	case DEPEND_PKG:
		d->w = NULL;
		break;

	case DEPEND_LIB:
		d->w = add_which_entry(&dl->wl, WHICH_LIBRARY, d->file);
		break;
			
	case DEPEND_RUN:
		if ((strrchr(d->file, '/') != NULL)
		  || (isnode(d->file) && isfile(d->file)))
			d->w = add_which_entry(&dl->wl, WHICH_FILE, d->file);
		else
			d->w = add_which_entry(&dl->wl, WHICH_EXEC, d->file);
		break;

	default:
		free(d);
		msg_error(2, "%s: invalid depend type.", __func__);
		return NULL;
	}
	PLIST_INSERT_TAIL(dl, d);
	return d;
}

/*
 * Check a dependency, setting d->pkg to the correct value.
 */
boolean
check_depend(depend_entry *d)
{

	switch(d->type) {
		case DEPEND_PKG:
			break;
			
		case DEPEND_LIB:
		case DEPEND_RUN:
			if (d->w)
				scan_which_entry(d->w);
			if (d->w && strlen(d->w->pkg) > 0)
				strlcpy(d->pkg, d->w->pkg, PATH_MAX);	  
			break;
	}
	if (strlen(d->pkg) == 0) {
		msg_warn("No package name to depend on! (%s)", (d->w ? d->file : "pkgdep"));
		/* We return true because we found the file */
		return TRUE;
	}
	if (pkg_mgr_check(d->pkg))
		return TRUE;
	return FALSE;
}

/*
 * Allocate a new require list entry.
 */
require_entry *
new_require_entry(void)
{
	require_entry *r;

	if ((r = (require_entry *)malloc(sizeof(require_entry))) == NULL) {
		msg_error(2, "%s: unable to allocate memory", __func__);
		return NULL;
	}	
	bzero(r, sizeof(require_entry));
	return r;
}

/*
 * Free an entire require list.
 */
void
free_require(require_list *rl)
{
	require_entry *r;

	if (rl == NULL)
		return;
	while (!PLIST_EMPTY(rl)) {
		PLIST_REMOVE_HEAD(rl, r);
		free(r);
	}
}

/*
 * Add an item to a require list.  We don't accept duplicates.
 */
require_entry *
add_require_entry(require_list *rl, char *pkg)
{
	require_entry *r;
	
	if (rl == NULL || pkg == NULL || strlen(pkg) == 0)
		return NULL;
	PLIST_FOREACH(r, rl) {
		if (!strcmp(r->pkg, pkg))
			return r;
	}
	if ((r = new_require_entry()) == NULL)
		return NULL;
	strlcpy(r->pkg, pkg, PATH_MAX);
	PLIST_INSERT_TAIL(rl, r);
	return r;
}

/*
 * Delete an an item from a require list.
 */
error_code
delete_require_entry(require_list *rl, char *pkg)
{
	require_entry *r, *tmp = NULL;
	
	if (rl == NULL || pkg == NULL || strlen(pkg) == 0)
		return FAIL;
	PLIST_FOREACH(r, rl) {
		if (strcmp(r->pkg, pkg) == 0) {
			PLIST_REMOVE(rl, r, tmp);
			free(r);
			r = tmp;
		}
	}
	return SUCCESS;
}
