/*
 * XXX: (reg) insert final project name here.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Jeremy D. Lea
 * XXX: (reg) Insert date
 *
 * Based on pkg_install, by:
 *   Jordan K. Hubbard
 *   18 July 1993
 *
 * Package entry handling routines.
 */

#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <ctype.h>
#include <dirent.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <md5.h>
#include <fetch.h>
#include <paths.h>
#include <regex.h>
#include "list.h"
#include "lib.h"

/*
 * Information for each package is divided between two structures.  struct
 * package contains the information which is obtained from the package, and
 * pkg_entry contains the system specfic information.
 */

/*
 * Allocate a new pkg_entry.
 */
pkg_entry *
new_pkg_entry(void)
{
	pkg_entry *pe;

	if ((pe = (pkg_entry *)malloc(sizeof(pkg_entry))) == NULL) {
		msg_error(2, "%s: unable to allocate memory", __func__);
		return NULL;
	}	
	bzero(pe, sizeof(pkg_entry));
	return pe;
}

/*
 * Orderly removal of our working structures.
 */
void
cleanup_pkg_entry(pkg_entry *pe)
{
	boolean force = PkgMgr.force;
	
	if (pe == NULL)
		return;
	if (pe->pkg != NULL) {
		/* If we crashed during install, try to delete the package. */
		if (pe->status == PKG_INSTALL) {
			PkgMgr.force = TRUE;
			delete_package(pe->pkg);
			PkgMgr.force = force;
		}
		free_package(pe->pkg);
		pe->pkg = NULL;
	}
	/* Do we want to leave the tmpdir behind? */
	if (pe->status != PKG_LEAVE) {
		remove_dir(pe->tmpdir);
		pe->tmpdir[0] = '\0';
	}
	/*
	 * Never leave a dangling dbdir, but don't commit any outstanding
	 * changes.
	 */
	if (pe->status != PKG_REGISTERED && pe->status != PKG_CHANGED) {
		remove_dir(pe->dbdir);
		pe->dbdir[0] = '\0';
		pe->status = PKG_UNKNOWN;
	}
}

/*
 * Allocate a new pkg_entry, and add it to the list.
 */
pkg_entry *
add_pkg_entry(void)
{
	pkg_entry *pe;
	
	if ((pe = new_pkg_entry()) == NULL)
		return NULL;
	PLIST_INSERT_TAIL(&PkgMgr, pe);
	return pe;
}

/*
 * Map a package name to a package entry.
 */
pkg_entry *
find_pkg_entry(const char *pkgname)
{
	pkg_entry *pe;
	char tmp[PATH_MAX];
	
	if (pkgname == NULL)
		return NULL;
	if (pkgname_of(tmp, pkgname, PATH_MAX) == NULL)
		return NULL;
	PLIST_FOREACH(pe, &PkgMgr) {
		/* Match package name before trying the dbdir */
		if (pe->pkg != NULL && (strlen(pe->pkg->name) > 0)) {
			if (strcmp(pe->pkg->name, tmp) == 0)
				return pe;
		} else if (strlen(pe->dbdir) > 0) {
			if (strstr(pe->dbdir, tmp) != NULL)
				return pe;
		}
	}
	return NULL;
}

/*
 * Return a list of packages matching the regex.  We append to
 * the list if it's not empty.
 */
error_code
match_pkg_entry(match_list *ml, const char *pkgregex)
{
	int err;
	char errbuf[PATH_MAX], pkgname[PATH_MAX], tmp[PATH_MAX];
	regex_t re;
	match_entry *m;
	pkg_entry *pe;
	boolean matched = FALSE;

	if (ml == NULL || pkgregex == NULL)
		return FAIL;
	pkgname_of(pkgname, pkgregex, PATH_MAX);
	if ((err = regcomp(&re, pkgregex, REG_BASIC | REG_NOSUB)) != 0) {
		regerror(err, &re, errbuf, sizeof(errbuf));
		msg_warn("%s: %s", pkgregex, errbuf);
		regfree(&re);
		return FAIL;
	}
	PLIST_FOREACH(pe, &PkgMgr) {
		/* Match package name before trying the dbdir */
		if (pe->pkg && (strlen(pe->pkg->name) > 0))
			pkgname_of(tmp, pe->pkg->name, PATH_MAX);
		else if (strlen(pe->dbdir) > 0)
			pkgname_of(tmp, pe->dbdir, PATH_MAX);
		else
			return FAIL;
		if ((err = regexec(&re, tmp, 0, NULL, 0)) == 0 ||
		  strcmp(tmp, pkgname) == 0) {
			/* XXX: (reg) we need a match.c */
			m = (match_entry *)malloc(sizeof(match_entry));
			bzero(m, sizeof(match_entry));
			m->pe = pe;
			PLIST_INSERT_TAIL(ml, m);
			matched = TRUE;
		} else if (err != REG_NOMATCH) {
			regerror(err, &re, errbuf, sizeof(errbuf));
			msg_warn("%s: %s", tmp, errbuf);
			regfree(&re);
			return FAIL;
		}
	}
	regfree(&re);
	if (!matched)
		return FAIL;
	return SUCCESS;
}

/*
 * Fetch a package, and extract it into it's tmpdir.
 */
error_code
fetch_pkg_entry(pkg_entry *pe, char *pkgname)
{
	char *cp = NULL;
	struct stat sb;

	if (pe == NULL || pkgname == NULL || strlen(pkgname) == 0)
		return FAIL;
	if (isURL(pkgname)) {
		if (make_tmpdir(pe->tmpdir, 0) == FAIL) {
			msg_warn("unable to make temp dir for FTP");
			return FAIL;
		}
		snprintf(pe->url, PATH_MAX, "%s", pkgname);
		pe->status = PKG_FETCH;
		if (fetch_package(pe->url) == FAIL) {
			msg_warn("unable to fetch `%s' by URL", pe->url);
			return FAIL;
		}
		pe->status = PKG_EXTRACTED;
	} else {
		if (strcmp(pkgname, "-") != 0) {
			/* expand all pathnames to fullnames */
			if ((cp = find_file(NULL, pkgname)) == NULL) {
				msg_warn("can't find package '%s'", pkgname);
				return FAIL;
			}
			strlcpy(pe->url, cp, PATH_MAX);
			free(cp);
			if (stat(pe->url, &sb) == FAIL) {
				msg_warn("can't stat package file '%s'", pe->url);
				return FAIL;
			}
			if (make_tmpdir(pe->tmpdir, sb.st_size * 4) == FAIL) {
				msg_warn("unable to make temp dir for %qd bytes", sb.st_size * 4);
				return FAIL;
			}
			if (unpack_package(pe->url, SPECIAL_FNAME) == FAIL) {
				msg_warn("unable to extract special files from `%s' - not a package?", pe->url);
				return FAIL;
			}
			pe->status = PKG_SPECIAL;
		} else {
			if (make_tmpdir(pe->tmpdir, 0) == FAIL) {
				msg_warn("unable to make temp dir");
				return FAIL;
			}
			if (unpack_package(pkgname, NULL) == FAIL) {
				msg_warn("unable to extract files from `%s' - not a package?", pkgname);
				return FAIL;
			}
			pe->status = PKG_EXTRACTED;
		}
	}
	return scan_pkg_entry(pe);
}

/*
 * Create a package from the installed special files, either in the tmpdir
 * or dbdir.  We stay cd'ed to the dir.
 */
error_code
scan_pkg_entry(pkg_entry *pe)
{
	int err_cnt = 0;
	
	if (pe == NULL)
		return FAIL;
	switch (pe->status) {
	case PKG_LEAVE:
	case PKG_FAIL:
	case PKG_UNKNOWN:
	case PKG_FETCH:
		msg_error(2, "package entry to immature to scan");
		return FAIL;
	
	case PKG_SPECIAL:
	case PKG_EXTRACTED:
	case PKG_TEMP:
	case PKG_INSTALL:
	case PKG_INSTALLED:
		if (strlen(pe->tmpdir) == 0) {
			msg_warn("package entry has no tmpdir");
			return FAIL;
		}
		if (enter_dir(pe->tmpdir) == FAIL)
			return FAIL;
		break;

	case PKG_REGISTERED:
	case PKG_CHANGED:
		if (strlen(pe->dbdir) == 0) {
			msg_warn("package entry has no dbdir");
			return FAIL;
		}
		if (enter_dir(pe->dbdir) == FAIL)
			return FAIL;
		break;

	case PKG_DELETE:
		msg_error(2, "package entry cannot be scaned during delete");
		return FAIL;
	}
	if (!isnode(CONTENTS_FNAME)) {
		msg_warn("package %s has no CONTENTS file!", pe->url);
		err_cnt++;
	}
	if (!isnode(COMMENT_FNAME)) {
		msg_warn("package %s has no COMMENT file!", pe->url);
		err_cnt++;
	}
	if (!isnode(DESC_FNAME)) {
		msg_warn("package %s has no DESC file!", pe->url);
		err_cnt++;
	}
	if (err_cnt > 0)
		return FAIL;
	if (pe->pkg == NULL) {
		if ((pe->pkg = new_package()) == NULL)
			return FAIL;
		if (read_package(pe->pkg) == FAIL)
			return FAIL;
		snprintf(pe->dbdir, PATH_MAX, "%s/%s", PkgMgr.dbdir, pe->pkg->name);
	}
	return SUCCESS;
}

/*
 * Extract a package, if it's not already extracted.
 */
error_code
extract_pkg_entry(pkg_entry *pe, boolean inplace)
{
	plist_entry *p;
	
	if (pe == NULL)
		return FAIL;
	/* We at least need a package to extract */
	if (pe->status < PKG_SPECIAL)
		return FAIL;
	/* It's already been extracted */
	if (pe->status > PKG_SPECIAL)
		return SUCCESS;
	if (scan_pkg_entry(pe) == FAIL)
		return FAIL;

	/* Extract directly rather than moving?  Oh goodie! */
	if (inplace) {
		msg_verbose("Doing in-place extraction for %s\n", pe->url);
		if ((p = find_plist_entry(&pe->pkg->plist, PLIST_CWD, NULL)) != NULL) {
			/* If this is a direct extract and we didn't want it, stop now. */
			if (PkgMgr.fake)
				return SUCCESS;
			if (!isdir(p->name)) {
				msg_verbose("Desired prefix of %s does not exist, creating..\n", p->name);
				make_hierarchy(p->name, NULL, NULL, NULL);
				if (enter_dir(p->name) == FAIL) {
					warn("unable to change directory to `%s'", p->name);
					return FAIL;
				}
			}
		} else {
			msg_warn("no prefix specified in `%s' - this is a bad package!", pe->url);
			return FAIL;
		}
		/* Finally unpack the whole mess */
		if (unpack_package(pe->url, NULL)) {
			msg_warn("unable to extract `%s'!", pe->url);
			return FAIL;
		}
	} else {
		if (enter_dir(pe->tmpdir) == FAIL)
			return FAIL;
		if (unpack_package(pe->url, NULL) == FAIL) {
			msg_warn("unable to extract `%s'!", pe->url);
			return FAIL;
		}
	}
	pe->status = PKG_EXTRACTED;
	return SUCCESS;
}

/*
 * Installs the package into prefix.
 */
error_code
install_pkg_entry(pkg_entry *pe)
{
	char *cp;
	plist_entry *p;

	if (pe == NULL)
		return FAIL;
	/* We at least need a package to extract */
	if (pe->status < PKG_SPECIAL)
		return FAIL;
	/* It's already been installed */
	if (pe->status > PKG_INSTALL)
		return SUCCESS;
	if (scan_pkg_entry(pe) == FAIL)
		return FAIL;
	if (pe->status == PKG_SPECIAL) {
		if (extract_pkg_entry(pe, (pe->pkg->options & PKG_OPTION_INPLACE
		                            ? TRUE : FALSE)) == FAIL)
			return FAIL;
	}
	/* See if we're already registered */
	if (pkg_mgr_check(pe->pkg->name) && !PkgMgr.force) {
		msg_warn("package `%s' already recorded as installed", pe->pkg->name);
		return FAIL;
	}
	if (depend_package(pe->pkg, pe->url) == FAIL)
		return FAIL;
	pe->status = PKG_INSTALL;
	if (install_package(pe->pkg) == FAIL)
		return FAIL;
	pe->status = PKG_INSTALLED;
	/* Time to record the deed? */
	if (!PkgMgr.norecord && !PkgMgr.fake) {
		if (register_pkg_entry(pe) == FAIL)
			return FAIL;
	}
	if ((p = find_plist_entry(&pe->pkg->plist, PLIST_DISPLAY, NULL)) != NULL) {
		if ((cp = read_file(p->name)) == NULL)
			msg_warn("cannot open %s as display file", p->name);
		else {
			printf("%s\n", cp);
			free(cp);
		}
	}
	return SUCCESS;
}

/*
 * Register a package entry into the dbdir.
 */
error_code
register_pkg_entry(pkg_entry *pe)
{

	if (pe == NULL)
		return FAIL;
	if (pe->status != PKG_INSTALLED)
		return FAIL;
	if (scan_pkg_entry(pe) == FAIL)
		return FAIL;
	if (getuid() != 0)
		msg_warn("not running as root - trying to record install anyway");
	msg_verbose("Attempting to record package into %s/.\n", pe->dbdir);
	if (make_hierarchy(pe->dbdir, "a+rx", NULL, NULL)) {
		msg_warn("can't make directory '%s', you're on your own!",
			pe->dbdir);
		return FAIL;
	}
	msg_verbose("Attempting to record package into %s/.\n", pe->dbdir);
	if (register_package(pe->pkg, pe->dbdir) == FAIL)
		return FAIL;
	if (add_package_filedb(pe->pkg) == FAIL)
		return FAIL;
	msg_verbose("Package %s registered in %s\n", pe->pkg->name, pe->dbdir);
	pe->status = PKG_REGISTERED;
	return SUCCESS;
}

/*
 * Rewrites the package information for an already registered package.
 */ 
error_code
write_pkg_entry(pkg_entry *pe)
{
	/* Just pretend to write */
	if (pe->status != PKG_CHANGED)
		return SUCCESS;
	/* If something changed, we can't have a NULL pkg. */
	if (pe->pkg == NULL)
		return FAIL;
	if (scan_pkg_entry(pe) == FAIL)
		return FAIL;
	if (write_package(pe->pkg) == FAIL)
		return FAIL;
	pe->status = PKG_REGISTERED;
	return SUCCESS;
}

/*
 * Upgrade one packge entry to another.
 */
error_code
upgrade_pkg_entry(pkg_entry *from, pkg_entry *to)
{

	if (from == NULL || to == NULL)
		return FAIL;
	if (from->status != PKG_REGISTERED)
		return FAIL;
	if (scan_pkg_entry(from) == FAIL
	  || scan_pkg_entry(to) == FAIL)
		return FAIL;
	if (to->status < PKG_SPECIAL || to->status > PKG_EXTRACTED)
		return FAIL;
	if (upgrade_package(from->pkg, to->pkg) == FAIL)
		return FAIL;
	if (delete_pkg_entry(from) == FAIL)
		return FAIL;
	if (install_pkg_entry(to) == FAIL)
		return FAIL;
	return SUCCESS;
}

/*
 * Delete a package entry.
 */
error_code
delete_pkg_entry(pkg_entry *pe)
{

	if (pe == NULL)
		return FAIL;
	if (pe->status != PKG_REGISTERED || strlen(pe->dbdir) == 0)
		return FAIL;
	if (scan_pkg_entry(pe) == FAIL)
		return FAIL;
	if (required_package(pe->pkg) && !PkgMgr.force)
		return FAIL;
	pe->status = PKG_DELETE;
	if (remove_package_filedb(pe->pkg) == FAIL && !PkgMgr.force)
		return FAIL;
	if (delete_package(pe->pkg) == FAIL)
		return FAIL;
	if (!PkgMgr.fake) {
		if (remove_dir(pe->dbdir)) {
			msg_warn("couldn't remove packge entry in %s, deinstall failed", pe->dbdir);
			if (!PkgMgr.force)
				return FAIL;
		}
	}
	return SUCCESS;
}

/*
 * Creates a tmpdir filled with the special files supplied.  The PkgMgr
 * will cleanup if we fail.
 */
pkg_entry *
create_pkg_entry(pkg_info *pi, depend_list *dl, char *pkgname)
{
	int err = 0;
	char tmp[PATH_MAX];
	pkg_entry *pe;
	
	if (pi == NULL || dl == NULL || pkgname == NULL)
		return NULL;
	/* Sanity check */
	if (pkgname_of(tmp, pkgname, PATH_MAX) == NULL) {
		msg_warn("required package name is missing");
		return NULL;
	}
	if (pi->contents == NULL || strlen(pi->contents) == 0) {
		msg_warn("required package contents list is missing");
		err++;
	}
	if (pi->comment == NULL || strlen(pi->comment) == 0) {
		msg_warn("required package comment string is missing");
		err++;
	}
	if (pi->desc == NULL || strlen(pi->comment) == 0) {
		msg_warn("required package description string is missing");
		err++;
	}
	if (err > 0)
		return NULL;
	if ((pe = add_pkg_entry()) == NULL)
		return NULL;;
	if (make_tmpdir(pe->tmpdir, 0) == FAIL) {
		msg_error(2, "unable to make temp dir for create");
		return NULL;
	}
	/* We'll crash if this isn't the case, with a bunch of unfound files */
	pe->status = PKG_INSTALLED;
	if ((pe->pkg = create_package(pi, dl, pkgname)) == NULL) {
		msg_error(2, "failed to create package!");
		return NULL;
	}
	if (scan_pkg_entry(pe) == FAIL) {
		msg_warn("failed to rescan package!");
		return NULL;
	}
	return pe;
}

/*
 * Packages installed files into a package.  pkgname can either be a file
 * name or a directory.
 */
error_code
package_pkg_entry(pkg_entry *pe, char *pkgname)
{
	char tmp[PATH_MAX];
	
	if (pe == NULL || pkgname == NULL)
		return FAIL;
	if (scan_pkg_entry(pe) == FAIL)
		return FAIL;
	if (isdir(pkgname))
		snprintf(tmp, PATH_MAX, "%s/%s%s", pkgname, pe->pkg->name, PKG_SUFFIX);
	else if (pkgname[0] != '/')
		snprintf(tmp, PATH_MAX, "%s/%s", PkgMgr.wrkdir, pkgname);
	else
		snprintf(tmp, PATH_MAX, "%s", pkgname);
	if (make_package(pe->pkg, tmp) == FAIL)
		return FAIL;
	return SUCCESS;
}

/*
 * Output pretty information about a package entry.  XXX: (reg) This should
 * be in show.c.
 */
error_code
show_pkg_entry(pkg_entry *pe, unsigned int flags, which_list *wl)
{
	char tmp[PATH_MAX];
	
	/*
	 * Index and which are special info types that have to override all
	 * others to make any sense. They also only apply to registered
	 * pacakges.
	 */
	if (flags & PKG_SHOW_INDEX) {
		if (pe->status != PKG_REGISTERED)
			return FAIL;
		if (enter_dir(pe->dbdir) == FAIL)
			return FAIL;
		pkgname_of(tmp, pe->dbdir, PATH_MAX);
		if (show_index(tmp, COMMENT_FNAME) == FAIL)
			return FAIL;
		return SUCCESS;
	}
	if (scan_pkg_entry(pe) == FAIL)
		return FAIL;
	if (flags & PKG_SHOW_WHICH) {
		if (pe->status != PKG_REGISTERED)
			return FAIL;
		if (show_which(pe->pkg, wl) == FAIL)
			return FAIL;
		return SUCCESS;
	}
	return show_package(pe->pkg, flags);
}

/*
 * Try to work out the version of a package entry compared to the latest
 * version.
 */
error_code
version_pkg_entry(pkg_entry *pe)
{
	if (pe == NULL)
		return FAIL;
	if (scan_pkg_entry(pe) == FAIL)
		return FAIL;
	return version_package(pe->pkg);
}
