/*
 * XXX: (reg) insert final project name here.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Jeremy D. Lea
 * XXX: (reg) Insert date
 *
 * Based on pkg_install, by:
 *   Jordan K. Hubbard
 *   18 July 1993
 *
 * Miscellaneous system routines.
 */

#include <sys/param.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "list.h"
#include "lib.h"

/*
 * Using fmt, replace all instances of:
 *
 * %F		With the parameter "name"
 * %D		With the parameter "dir"
 * %B		Return the directory part ("base") of %D/%F
 * %f		Return the filename part of %D/%F
 */
error_code
format_cmd(char *buf, int max, const char *fmt, const char *dir, const char *name)
{
	int i = 0;
	char tmp[PATH_MAX];

	while (fmt[0] != '\0' && i < (max-1)) {
		buf[i] = '\0'; 
		if (fmt[0] == '%') {
			switch (*++fmt) {
			case 'F':
				i = strlcat(buf, name, max);
				break;
			case 'D':
				i = strlcat(buf, dir, max);
				break;
			case 'B':
				strlcat(buf, dir, max);
				strlcat(buf, "/", max);
				i = strlcat(buf, name, max);
				if (i > (max-1))
					break;
				while (i > 0 && buf[i] != '/')
					i--;
				buf[i] = '\0';
				break;
			case 'f':
				snprintf(tmp, max, "%s/%s", dir, name);
				i = strlcat(buf, basename_of(tmp), max);
				break;
			default:
				buf[i++] = fmt[0];
				break;
			}
			++fmt;
		} else
			buf[i++] = *fmt++;
	}
	if (i >= (max-1)) {
		msg_error(2, "%s: tried to format too long a command", __func__);
		buf[0] = '\0';
		return FAIL;
	}
	buf[i] = '\0';
	return SUCCESS;
}

/*
 * Unusual system() substitute.  Accepts format string and args,
 * builds and executes command.  Returns exit code.
 */
error_code
vsystem(const char *fmt, ...)
{
	va_list args;
	char *cmd;
	int err, maxargs;

	if (fmt == NULL || strlen(fmt) == 0) {
		msg_error(2, "%s: tried to exec a null command", __func__);
		return FAIL;
	}
	/* Leave some slop for the sh -c. */
	maxargs = sysconf(_SC_ARG_MAX) - 32;
	if ((cmd = malloc(maxargs)) == NULL) {
		msg_error(2, "%s: can't alloc arg space", __func__);
		return FAIL;
	}
	va_start(args, fmt);
	if (vsnprintf(cmd, maxargs, fmt, args) > maxargs) {
		msg_error(2, "%s: args are too long", __func__);
		return FAIL;
	}
#ifdef DEBUG
	printf("Executing %s\n", cmd);
#endif
	fflush(NULL);
	err = system(cmd);
	va_end(args);
	free(cmd);

	return err;
}

/*
 * Unusual popen() substitute.  Accepts format string and args, builds and
 * executes command.  The caller must free the returned buffer.
 */
char *
vpipe(const char *fmt, ...)
{
	va_list args;
	char *cmd, *rp;
	int maxargs;
	FILE *fp;

	if (fmt == NULL || strlen(fmt) == 0) {
		msg_error(2, "%s: tried to exec a null command", __func__);
		return NULL;
	}
	if ((rp = malloc(PATH_MAX)) == NULL) {
		msg_error(2, "%s: can't alloc output buffer space", __func__);
		return NULL;
	}
	/* Leave some slop for the sh -c. */
	maxargs = sysconf(_SC_ARG_MAX) - 32;
	if ((cmd = malloc(maxargs)) == NULL) {
		msg_error(2, "%s: can't alloc arg space", __func__);
		return NULL;
	}
	va_start(args, fmt);
	if (vsnprintf(cmd, maxargs, fmt, args) > maxargs) {
		msg_error(2, "%s: args are too long", __func__);
		return NULL;
	}
#ifdef DEBUG
	printf("Executing %s\n", cmd);
#endif
	fflush(NULL);
	fp = popen(cmd, "r");
	read_line(rp, PATH_MAX, fp);
#ifdef DEBUG
	printf("Returned %s\n", rp);
#endif
	va_end(args);
	free(cmd);
	if (pclose(fp) || (strlen(rp) == 0)) {
		free(rp);
		return NULL;
	}
	return rp;
}
