/*
 * XXX: (reg) insert final project name here.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Jeremy D. Lea
 * XXX: (reg) Insert date
 *
 * Based on pkg_install, by:
 *   Jordan K. Hubbard
 *   18 July 1993
 *
 * Miscellaneous file access utilities.
 */

#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/signal.h>
#include <sys/mount.h>
#include <md5.h>
#include "list.h"
#include "lib.h"

/*
 * Quick check to see if a file exists.
 */
boolean
isnode(const char *fname)
{
	struct stat sb;

	if (lstat(fname, &sb) != FAIL)
		return TRUE;
	return FALSE;
}

/*
 * Quick check to see if something is a directory or symlink to a
 * directory.
 */
boolean
isdir(const char *fname)
{
	char tmp[PATH_MAX];
	struct stat sb;

	if (lstat(fname, &sb) != FAIL && S_ISDIR(sb.st_mode))
		return TRUE;
	snprintf(tmp, PATH_MAX, "%s/", fname);
	if (lstat(tmp, &sb) != FAIL && S_ISDIR(sb.st_mode))
		return TRUE;
	return FALSE;
}

/*
 * Check to see if file is a dir or symlink to a dir, and is empty.
 */
boolean
isemptydir(const char *fname)
{
	DIR *dirp;
	struct dirent *dp;

	if (!isdir(fname))
		return FALSE;
	/* Check to see if we can open the dir.  If no perms, leave it. */
	if ((dirp = opendir(fname)) == NULL)
		return FALSE;
	for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp)) {
		if (strcmp(dp->d_name, ".") && strcmp(dp->d_name, "..")) {
			closedir(dirp);
			return FALSE;
		}
	}
	closedir(dirp);
	return TRUE;
}

/*
 * Returns TRUE if file is a regular file or symlink pointing to a regular
 * file.
 */
boolean
isfile(const char *fname)
{
	struct stat sb;

	if (stat(fname, &sb) != FAIL && S_ISREG(sb.st_mode))
		return TRUE;
	return FALSE;
}

/* 
 * Check to see if file is a file or symlink pointing to a file and is
 * empty.  If nonexistent or not a file, pretend it's empty and return TRUE
 * as if zero sized.
 */
boolean
isemptyfile(const char *fname)
{
	struct stat sb;

	if (stat(fname, &sb) != FAIL && S_ISREG(sb.st_mode)) {
		if (sb.st_size != 0)
			return FALSE;
	}
	return TRUE;
}

/*
 * Returns TRUE if file is a symbolic link.
 */
boolean
issymlink(const char *fname)
{
	struct stat sb;

	if (lstat(fname, &sb) != FAIL && S_ISLNK(sb.st_mode))
		return TRUE;
	return FALSE;
}


/*
 * Returns TRUE if file is a URL specification.  We just handle ftp, http
 * and file at the moment.  This should be a function in libfetch.
 */
boolean
isURL(const char *fname)
{
	const char *URLs[] = { "ftp://", "http://", "file://", NULL };
	int i = -1;

	if (fname == NULL)
		return FALSE;
	while (isspace(*fname))
		fname++;
	while (URLs[++i] != NULL) {
		if (strncmp(fname, URLs[i], strlen(URLs[i])) == 0)
			return TRUE;
	}
	return FALSE;
}

/*
 * Try to convert a known good package location into a location for a
 * dependency.  Caller must free the returned string.
 */
char *
find_file(const char *base, const char *fname)
{
	char tmp[PATH_MAX];
	char *cp, *cp2, *rp;
	const char *suffix;

	if (fname == NULL)
		return NULL;
	/* 
	 * There's not much point in figuring out the right size for the buffer,
	 * so just make it big enough.
	 */
	if ((rp = malloc(PATH_MAX)) == NULL)
		return NULL;
	suffix = strrchr(fname, '.');
	/* XXX: (reg) bzip packages */
	if (suffix == NULL || (strcmp(suffix, PKG_SUFFIX)
	  && strcmp(suffix, ".tgz") && strcmp(suffix, ".tar")
	  && strcmp(suffix, ".Z") && strcmp(suffix, ".gz")))
		suffix = PKG_SUFFIX;
	else
		suffix = "";
	strlcpy(tmp, fname, PATH_MAX);
	strlcat(tmp, suffix, PATH_MAX);
	if (isURL(tmp))
		return strcpy(rp, tmp);
	if (isfile(tmp))
		return realpath(tmp, rp);
	/* We've not found the file yet, so store the base filename */
	strlcpy(rp, basename_of(fname), PATH_MAX);
	strlcat(rp, suffix, PATH_MAX);
	if (base) {
		/*
		 * We've been given an existing package (that's known to be good)
		 * and now we need to construct a composite one out of that and the
		 * package name we were handed as a dependency.
		 *
		 * This has two code paths, one for files and one for URLs.
		 */
		strlcpy(tmp, base, PATH_MAX);
		/* First try to chop off the basename of base and add our filename */
		if ((cp = strrchr(tmp, '/')) != NULL) {
			cp[1] = '\0';
			if (!isURL(tmp)) {
				strlcat(tmp, rp, PATH_MAX);
				if (isfile(tmp))
					return realpath(tmp, rp);
			}
			/* Go up one more dir (the category) */
			cp[0] = '\0';
		}
		/* We suspect this might be a URL, so look in the All directory */
		if ((cp = strrchr(tmp, '/')) != NULL) {
			cp[1] = '\0';
			strlcat(tmp, "All/", PATH_MAX);
			strlcat(tmp, rp, PATH_MAX);
			if (isURL(tmp)) {
				return strcpy(rp, tmp);
			}
			if (isfile(tmp))
				return realpath(tmp, rp);
		}
	}

	/* Still no luck, so start looking in our paths */
	if (PkgMgr.wrkdir != NULL) {
		snprintf(tmp, PATH_MAX, "%s/%s", PkgMgr.wrkdir, rp);
		if (isfile(tmp))
			return realpath(tmp, rp);
	}
	cp = getenv(PKG_PATH);
	while (cp != NULL) {
		cp2 = strsep(&cp, ":");
		snprintf(tmp, PATH_MAX, "%s/%s", cp2 ? cp2 : cp, rp);
		if (isfile(tmp))
			return realpath(tmp, rp);
	}
	free(rp);
	return NULL;
}

/*
 * Get a fixed length string from a file, NULL terminated.  We always gobble
 * up the full line, and any preceeding empty lines.
 */
char *
read_line(char *str, int max, FILE *fp)
{
	char *cp;
	size_t len;
	
	if (str == NULL || fp == NULL)
		return NULL;
	str[0] = '\0';
	while ((cp = fgetln(fp, &len)) != NULL) {
		strncpy(str, cp, MIN(len, max));
		str[MIN(len, max)-1] = '\0';
		len = strlen(str);
		while (len > 0 && isspace(str[len-1]))
			str[--len] = '\0';
		if (len > 0)
			return str;
	}
	return NULL;
}

/*
 * Read the contents of a file into a string.
 */
char *
read_file(const char *fname)
{
	char *rp;
	struct stat sb;
	FILE *fp;

	/* We don't use isfile() because we want the size. */
	if (stat(fname, &sb) != FAIL && S_ISREG(sb.st_mode)) {
		msg_error(2, "%s: '%s' isn't a file", __func__, fname);
		return NULL;
	}
	if ((rp = malloc(sb.st_size + 1)) == NULL) {
		msg_error(2, "%s: unable to allocate memory", __func__, fname);
		return NULL;
	}
	if ((fp = fopen(fname, "r")) == NULL) {
		msg_error(2, "%s: unable to fopen '%s' for reading", __func__, fname);
		free(rp);
		return NULL;
	}
	if (fread(rp, 1, sb.st_size, fp) != sb.st_size) {
		msg_error(2, "%s: short fread on '%s' - failed to read %qd bytes",
		  __func__, fname, sb.st_size);
		free(rp);
		fclose(fp);
		return NULL;
	}
	fclose(fp);
	rp[sb.st_size] = '\0';
	return rp;
}

/*
 * Write the contents of "str" to a file
 */
error_code
write_file(const char *fname, const char *str)
{
	FILE *fp;
	int len;

	if ((fp = fopen(fname, "w")) == NULL) {
		msg_error(2, "%s: cannot open '%s' for writing", __func__, fname);
		return FAIL;
	}
	len = strlen(str);
	if (fwrite(str, 1, len, fp) != len) {
		msg_error(2, "%s: short fwrite on '%s' - tried to write %qd bytes",
		  __func__, fname, len);
		return FAIL;
	}
	if (fclose(fp) == FAIL) {
		msg_error(2, "%s: failed to fclose '%s'", __func__, fname);
		return FAIL;
	}
	return SUCCESS;
}

/*
 * Copy a file.  If from is not an absolute path, then use the dir.
 */
error_code
copy_file(const char *dir, const char *from, const char *to)
{
	const char *cp;

	if (from[0] == '/') {
		cp = from + 1;
		dir = "";
	} else
		cp = from;
	if (vsystem("/bin/cp -r %s/%s %s", dir, from, to)) {
		/* XXX: (reg) error? */
		msg_warn("could not copy from '%s/%s' to '%s'", dir, from, to);
		return FAIL;
	}
	return SUCCESS;
}

/*
 * Move a file.  If from is not an absolute path, then use the dir.
 */
error_code
move_file(const char *dir, const char *from, const char *to)
{
	const char *cp;

	if (from[0] == '/') {
		cp = from + 1;
		dir = "";
	} else
		cp = from;
	if (vsystem("/bin/mv %s/%s %s", dir, from, to)) {
		/* XXX: (reg) error? */
		msg_warn("could not move from '%s/%s' to '%s'", dir, from, to);
		return FAIL;
	}
	return SUCCESS;
}

/*
 * Make a hard link.  Also try to make a directory for to.
 * XXX: (reg) not used.
 */
error_code
link_file(const char *from, const char *to)
{
	const char *cp;
	
	if (link(from, to) == FAIL) {
		if (errno == ENOENT) {
			/* try making the container directory */
			cp = strrchr(to, '/');
			if (cp != NULL)
				vsystem("/bin/mkdir -p %.*s", cp - to, to);
			return link(from, to);
		}
		return FAIL;
	}
	return SUCCESS;
}

/*
 * Apply permissions to file/dir.
 */
error_code
apply_perms(const char *dir, const char *file,
  const char *m, const char *o, const char *g)
{
	const char *cd_to;

	if (dir == NULL || file[0] == '/')		/* absolute path? */
		cd_to = "/";
	else
		cd_to = dir;
	if (m != NULL) {
		if (vsystem("cd %s && /bin/chmod -R %s %s", cd_to, m, file)) {
			msg_warn("couldn't change modes of '%s' to '%s'", file, m);
			return FAIL;
		}
	}
	if (o != NULL || g != NULL) {
		if (vsystem("cd %s && /usr/sbin/chown -R %s:%s %s", cd_to,
		  (o ? o : ""), (g ? g : ""), file)) {
			msg_warn("couldn't change owner/group of '%s' to '%s:%s'",
			  file, (o ? o : ""), (g ? g : ""));
			return FAIL;
		}
	}
	return SUCCESS;
}

/*
 * Assuming dir is a desired directory name, make it and all intervening
 * directories necessary.
 */
error_code
make_hierarchy(const char *dir, const char *m, const char *o, const char *g)
{
	char tmp[PATH_MAX];
	char *cp;

	if (dir == NULL)
		return FAIL;
	strlcpy(tmp, dir, PATH_MAX);
	cp = &tmp[strlen(tmp)];
	/* Zap all /'s at the end of dir. */
	while (--cp > tmp && cp[0] == '/')
		cp[0] = '\0';
	if (cp == tmp)
		return FAIL;
	cp = tmp;
	while (cp != NULL) {
		/* Move after the '/' */
		while (cp[0] != '\0' && cp[0] == '/')
			cp++;
		/* Zap the next '/' */
		if ((cp = strchr(cp, '/')) != NULL)
			cp[0] = '\0';
		if (isnode(tmp)) {
			if (!isdir(tmp))
				return FAIL;
		} else {
			if (vsystem("/bin/mkdir %s", tmp)) {
				msg_error(2, "%s: unable to make directory", __func__);
				return FAIL;
			}
			if (apply_perms(NULL, tmp, m, o, g) == FAIL)
				return FAIL;
		}
		/* Put it back */
		if (cp != NULL)
			cp[0] = '/';
	}
	return SUCCESS;
}

/* Usually "rm" and "rmdir", but often "echo" during debugging! */
#define	REMOVE(dir,ie) vsystem("/bin/rm -r%s %s", (ie ? "f" : ""), dir)
#define	RMDIR(dir,nuke) vsystem("/bin/rmdir %s%s", (nuke ? "-p " : ""), dir)

/*
 * Selectively delete a hierarchy.
 */
error_code
delete_hierarchy(const char *dir, boolean ign_err, boolean nukedirs)
{

	if (!isnode(dir)) {
		if (!ign_err)
			msg_warn("`%s' doesn't really exist", dir);
		return (ign_err ? SUCCESS : FAIL);
	} else if (nukedirs) {
		if (REMOVE(dir, ign_err) || (RMDIR(dir, nukedirs) && !ign_err))
			return FAIL;
	} else if (isdir(dir) && !issymlink(dir)) {
		if (RMDIR(dir, nukedirs) && !ign_err)
			return FAIL;
	} else {
		if (REMOVE(dir, ign_err))
			return FAIL;
	}
	return SUCCESS;
}

/*
 * Determine the free space available in a directory.
 */
off_t
min_free(const char *dir)
{
	struct statfs buf;

	if (statfs(dir, &buf) == FAIL) {
		msg_warn("statfs failed in %s", dir);
		return -1;
	}
	return (off_t)(buf.f_bavail*buf.f_bsize);
}

/*
 * Find a good place to play. tmpdir should be PATH_MAX long.
 */
error_code
make_tmpdir(char *tmpdir, size_t sz)
{
	char *cp[5];
	int i;
	struct stat sb;

	/*
	 * We don't use a NULL terminated array of char * here, because we
	 * want to check the two environment variables first, and getenv
	 * might return NULL;
	 */
	cp[0] = getenv(PKG_TMPDIR);
	cp[1] = getenv(SYS_TMPDIR);
	cp[2] = "/var/tmp";
	cp[3] = "/tmp";
	cp[4] = "/usr/tmp";
	for (i = 0; (i < 5) && (cp[i] == NULL || stat(cp[i], &sb) == FAIL
	            || min_free(cp[i]) < sz); i++)
		;
	if (i == 5) {
		msg_warn("can't find enough temporary space to extract the files, please set your PKG_TMPDIR environment variable to a location with at least %d bytes free", sz);
		return FAIL;
	}
	snprintf(tmpdir, PATH_MAX, "%s/%s", cp[i], PkgMgr.mktmp_template);
	if (mkdtemp(tmpdir) == NULL) {
		msg_warn("can't mktemp '%s'", tmpdir);
		return FAIL;
	}
	if (chmod(tmpdir, 0755) == FAIL) {
		msg_warn("can't chmod on '%s'", tmpdir);
		return FAIL;
	}
	if (PkgMgr.verbose && sz)
		fprintf(stderr, "Requested space: %d bytes, free space: %qd bytes in %s\n", (int)sz, min_free(tmpdir), tmpdir);
	return enter_dir(tmpdir);
}

/*
 * chdir() to an existing dir.  dir should be PATH_MAX long, and is
 * replaced with the new CWD.
 */
error_code
enter_dir(char *dir)
{

	if (dir == NULL || strlen(dir) == 0 || !isdir(dir))
		return FAIL;
	if (chdir(dir) == FAIL) {
		msg_warn("can't chdir to '%s'", dir);
		return FAIL;
	}
	getcwd(dir, PATH_MAX);
	return SUCCESS;
}

/* 
 * Remove a dir, changing back to our wrkdir if needed.
 */
error_code
remove_dir(const char *dir)
{
	char tmp[PATH_MAX];
	void (*oldsig)(int);

	if (dir == NULL || strlen(dir) == 0 || !isdir(dir))
		/* Just pretend. */
		return FAIL;
	/* Don't interrupt while we're cleaning up. */
	oldsig = signal(SIGINT, SIG_IGN);
	if (!getcwd(tmp, PATH_MAX) || strcmp(tmp, dir) == 0) {
		if (strlen(PkgMgr.wrkdir) > 0 && chdir(PkgMgr.wrkdir) == FAIL)
			msg_warn("couldn't chdir back to '%s'", PkgMgr.wrkdir);
	}
	if (REMOVE(dir, TRUE)) {
		signal(SIGINT, oldsig);
		msg_warn("couldn't remove temporary dir '%s'", dir);
		return FAIL;
	}
	signal(SIGINT, oldsig);
	return SUCCESS;
}

/*
 * Compute the MD5 checksum for a file, returning it in a buffer the caller
 * must free.
 */
char *
make_md5(const char *fname)
{
	char *rp;

	if (fname == NULL || strlen(fname) == 0)
		return NULL;
	if (!isnode(fname) || !isfile(fname))
		return NULL;
	if ((rp = malloc(33)) == NULL)
		return NULL;
	if (MD5File(fname, rp) == NULL) {
		free(rp);
		return NULL;
	}
	return rp;
}

/*
 * Check the MD5 checksum of a file.
 */
boolean
check_md5(const char *fname, const char *md5)
{
	char *cp, buf[33];

	if (md5 == NULL || strlen(md5) == 0)
		return FALSE;	
	if (fname == NULL || strlen(fname) == 0)
		return FALSE;
	if ((cp = MD5File(fname, buf)) == NULL)
		return FALSE;
	if (strcmp(cp, md5) != 0)
		return FALSE;
	return TRUE;
}
