/*
 * XXX: (reg) insert final project name here.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Jeremy D. Lea
 * XXX: (reg) Insert date
 *
 * Based on pkgdb.c, by Hubert Feyrer.
 *
 * Routines for handling a database of files and packages that installed
 * them.
 */

#include <db.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "list.h"
#include "lib.h"

/*
 * NetBSD's pkg_install has a DB based database of installed files.  This is
 * a slightly different implementation.  This uses some nastier tricks... 
 * Because we treat it more like a cache, we realpath(3) everything.  If the
 * machine's discs are shuffled, then we'll just end rebuilding the
 * database.  Because we rely so heavily on knowing who installed what, we
 * never trust the database, and blow it away if we so much as suspect it is
 * lying to us.
 *
 * XXX: (reg) we need some uid checks in here...
 */ 

/*
 * Just use a global database pointer, because we're lazy...
 */
static DB *filedb = NULL;

/*
 * We don't want just anyone grubbing around in the database...
 */
static error_code	open_filedb(boolean);
static void			close_filedb(void);
static error_code	rebuild_filedb(char *, BTREEINFO *);
static const char	*get_filedb(const char *);
static error_code	put_filedb(const char *, const char *);
static error_code	remove_filedb(const char *);

/*
 * Open the database.  If we fail, or if requested, we try to rebuild and
 * fail completely if we cannot.
 */
error_code
open_filedb(boolean rebuild)
{
	char dbfile[PATH_MAX];
	BTREEINFO info;

	if (filedb != NULL)
		close_filedb();
	info.flags = 0;
	info.cachesize = 2*1024*1024;
	info.maxkeypage = 0;
	info.minkeypage = 0;
	info.psize = 4096;
	info.compare = NULL;
	info.prefix = NULL;
	info.lorder = 0;
	snprintf(dbfile, PATH_MAX, "%s/%s", LOG_DIR, PKGDB_FILE);
	if (!rebuild)
		filedb = dbopen(dbfile, O_RDWR, 0644, DB_BTREE, &info);
	if (filedb == NULL) {
		if (rebuild_filedb(dbfile, &info) == FAIL);
			return FAIL;
		filedb = dbopen(dbfile, O_RDWR, 0644, DB_BTREE, &info);
		if (filedb == NULL)
			return FAIL;
	}
	return SUCCESS;
}

/*
 * Close the file database
 */
void
close_filedb(void)
{
	if (filedb != NULL)
		(filedb->close)(filedb);
	filedb = NULL;
}

/*
 * Find the value "pkgname" associated with the key "file" in the database.
 */
const char *
get_filedb(const char *file)
{
	int status;
	DBT key, val;

	if (filedb == NULL)
		return NULL;
	key.data = (void *)file;
	key.size = strlen(file) + 1;
	if ((status = (filedb->get)(filedb, &key, &val, 0)) == -1) {
		close_filedb();
		return NULL;
	}
	if (status == 1)	
		return NULL;
	return val.data;
}

/*
 * Fill a which_entry based on the database.  If the database lies to us, or
 * shows some error, we blow it away and retry.  If it fails again, we fail.
 */
error_code
which_filedb(which_entry *w)
{
	const char *cp;
	pkg_entry *pe;
	boolean retrying = FALSE;

	if (w == NULL || strlen(w->file) == 0)
		return FAIL;
	w->pkg[0] = '\0';
	if (open_filedb(FALSE) == FAIL)
		return FAIL;
	while (1) {
    	if ((cp = get_filedb(w->file)) != NULL) {
    		close_filedb();
    		if ((pe = find_pkg_entry(cp)) != NULL) {
    			if (scan_pkg_entry(pe))
    				return FAIL;
    			if (check_which(w, pe->pkg))
    				return SUCCESS;
    		}
    	}
    	if (filedb != NULL) {
			/* We really didn't find it. */
			close_filedb();
			return SUCCESS;
		}
		if (retrying || open_filedb(TRUE) == FAIL)
    		return FAIL;
		retrying = TRUE;
	}
}

/*
 * Rebuild the database.  We just blow it away, then loop through all of
 * the registered packages, adding all of the files.
 */
error_code
rebuild_filedb(char *dbfile, BTREEINFO *info)
{
	char tmp[PATH_MAX];
	pkg_entry *pe;
	error_code rv = SUCCESS;

	if (filedb != NULL)
		close_filedb();
	if (!getcwd(tmp, PATH_MAX))
		snprintf(tmp, PATH_MAX, PkgMgr.wrkdir);
	filedb = dbopen(dbfile, O_RDWR | O_CREAT | O_TRUNC, 0644, DB_BTREE, info);
	if (filedb != NULL) {
		close_filedb();
    	PLIST_FOREACH(pe, &PkgMgr) {
    		if (pe->status == PKG_REGISTERED) {
    			if (scan_pkg_entry(pe)) {
    				rv = FAIL;
    				break;
    			}
    			if ((rv = add_package_filedb(pe->pkg)) == FAIL)
    				break;
				/* XXX: (reg) do we need this? */
				cleanup_pkg_entry(pe);
    		}
    	}
	}
	enter_dir(tmp);
	if (rv == FAIL) {
		unlink(dbfile);
		msg_error(2, "Failed to rebuild %s", dbfile);
	}
	return rv;
}

/*
 * Store value "file" with key "pkgname" in database.  We just return an
 * error if the key already exists.  This should never happen, because we
 * always check first.
 */
error_code
put_filedb(const char *file, const char *pkgname)
{ 
	DBT key, val;

	if (filedb == NULL)
		return FAIL;
	key.data = (void *)file;
	key.size = strlen(file) + 1;
	val.data = (void *)pkgname;
	val.size = strlen(pkgname) + 1;
	if ((filedb->put)(filedb, &key, &val, R_NOOVERWRITE) != 0) {
		close_filedb();
		return FAIL;
	}
	return SUCCESS;
}

/*
 * Add all the files in a package to the database.
 */
error_code
add_package_filedb(package *pkg)
{
	plist_entry *p;
	char *dir = ".";
	char tmp[PATH_MAX], rp[PATH_MAX];
	const char *cp;

	if (open_filedb(FALSE) == FAIL)
		return FAIL;
	PLIST_FOREACH(p, &pkg->plist) {
		switch(p->type) {
		case PLIST_CWD:
			dir = p->name;
			break;
	
		case PLIST_FILE:
			if (p->ignore)
				break;
			snprintf(tmp, PATH_MAX, "%s/%s", dir, p->name);
			/* We only add real files to the database. */
			if (!isfile(tmp))
				break;
			realpath(tmp, rp);
			if ((cp = get_filedb(rp)) != NULL) {
				if (strcmp(cp, pkg->name) != 0)
					msg_warn("Both %s and %s claim to have installed %s.",
					  cp, pkg->name, rp);
			} else if (put_filedb(rp, pkg->name) == FAIL) {
				return FAIL;
			}
			break;
	
		default:
			break;
		}
	}
	close_filedb();
	return SUCCESS;
}

/*
 * Remove the record associated with the key "file" from the database.  If
 * the file is not listed, we return an error, because we always check
 * before we try to delete the file.
 */
error_code
remove_filedb(const char *file)
{
	DBT key;

	if (filedb == NULL)
		return FAIL;
	key.data = (void *)file;
	key.size = strlen(file) + 1;
	if ((filedb->del)(filedb, &key, 0) != 0) {
		close_filedb();
		return FAIL;
	}
	return SUCCESS;
}

/*
 * Remove all the files in a package from the database.
 */
error_code
remove_package_filedb(package *pkg)
{
	plist_entry *p;
	char *dir = ".";
	char tmp[PATH_MAX], rp[PATH_MAX];
	const char *cp;

	if (open_filedb(FALSE) == FAIL)
		return FAIL;
	PLIST_FOREACH(p, &pkg->plist) {
		switch(p->type) {
		case PLIST_CWD:
			dir = p->name;
			break;
	
		case PLIST_FILE:
			if (p->ignore)
				break;
			snprintf(tmp, PATH_MAX, "%s/%s", dir, p->name);
			if (!isfile(tmp))
				break;
			realpath(tmp, rp);
			if ((cp = get_filedb(rp)) == NULL) {
				if (filedb == NULL)
					return FAIL;
				else
					break;
			}
			if (strcmp(cp, pkg->name) != 0)
				msg_warn("Both %s and %s claim to have installed %s.",
				  cp, pkg->name, rp);
			else
				if (remove_filedb(rp) == FAIL)
					return FAIL;
			break;
	
		default:
			break;
		}
	}
	close_filedb();
	return SUCCESS;
}
