/*
 * XXX: (reg) insert final project name here.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Jeremy D. Lea
 * XXX: (reg) Insert date
 *
 * Based on pkg_install, by:
 *   Jordan K. Hubbard
 *   18 July 1993
 *
 * Package handling routines.  These only deal with localhost independent
 * parts (which is most of the package).
 */

#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <ctype.h>
#include <dirent.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <md5.h>
#include <fetch.h>
#include <paths.h>
#include "list.h"
#include "lib.h"

/*
 * Allocate a new package structure.
 */
package *
new_package(void)
{
	package *pkg;

	if ((pkg = (package *)malloc(sizeof(package))) == NULL) {
		msg_error(2, "%s: unable to allocate memory", __func__);
		return NULL;
	}	
	bzero(pkg, sizeof(package));
	return pkg;
}

/*
 * Orderly removal of our working structures.
 */
void
free_package(package *pkg)
{
	if (pkg == NULL)
		return;
	free_depend(&pkg->depends);
	free_require(&pkg->requires);
	free_plist(&pkg->plist);
	free(pkg);
}

#define	BUF_SIZE	8192

/*
 * Try and fetch a file by URL.
 */
error_code
fetch_package(char *fname)
{
	char buf[BUF_SIZE];
	FILE *ftp;
	pid_t tpid;
	int pfd[2], pstat;
	size_t r, w = 0;

	if (fname == NULL || !isURL(fname))
		return FAIL;
	if ((ftp = fetchGetURL(fname, NULL)) == NULL) {
		msg_warn("FTP Unable to get %s: %s", fname, fetchLastErrString);
		return FAIL;
	}
	msg_verbose("Fetching %s...", fname);
	/* So that libfetch's ouput appears in order... */
	fflush(stdout);
	if (pipe(pfd) == -1) {
		msg_error(2, "%s: pipe()", __func__);
		return FAIL;
	}
	if ((tpid = fork()) == -1) {
		msg_error(2, "$s: fork()", __func__);
		return FAIL;
	}
	if (tpid == 0) {
		close(pfd[1]);
		dup2(pfd[0], 0);
		close(pfd[0]);
		execl("/usr/bin/tar", "tar", PkgMgr.verbose ? "-xzvf" : "-xzf", "-", 0);
		_exit(2);
	}
	close(pfd[0]);
	for (;;) {
		if ((r = fread(buf, 1, BUF_SIZE, ftp)) < 1)
			break;
		if ((w = write(pfd[1], buf, r)) != r)
			break;
	}
	fclose(ftp);
	close(pfd[1]);
	if (r == -1) {
		msg_warn("warning: error reading from server");
		return FAIL;
	}
	if (w == -1) {
		msg_warn("warning: error writing to tar");
		return FAIL;
	}
	tpid = waitpid(tpid, &pstat, 0);
	msg_verbose("tar command returns %d status\n", WEXITSTATUS(pstat));
	msg_verbose(" Done.\n");
	return SUCCESS;
}

/*
 * Unpack a package, or just the files in flist.
 */
error_code
unpack_package(char *pkg, char *flist)
{
	char args[20];

	if (pkg == NULL || strlen(pkg) == 0)
		return FAIL;
	args[0] = '\0';
	/*
	 * Figure out by a crude heuristic whether this or not this is probably
	 * compressed.
	 */
	if (!strcmp(pkg, "-") || tolower(pkg[strlen(pkg) - 1]) == 'z')
		strcat(args, "-z");
	else if (tolower(pkg[strlen(pkg) - 1]) == '2')
		strcat(args, "-y");
	if (flist != NULL)
		strcat(args, " --fast-read"); 
	if (vsystem("/usr/bin/tar %s %s -xpf '%s'", args, flist ? flist : "", pkg)) {
		msg_warn("tar extract of %s failed!", pkg);
		return FAIL;
	}
	return SUCCESS;
}

/*
 * Creates a tmpdir filled with the special files supplied.
 */
package *
create_package(pkg_info *pi, depend_list *dl, char *pkgname)
{
	char *cp;
	int err = 0;
	boolean found;
	char prefix[PATH_MAX], tmp[PATH_MAX], pline[PATH_MAX];
	FILE *fp;
	package *pkg;
	plist_entry *p;
	depend_entry *d;
	plist_t type;
	
	if (pi == NULL || dl == NULL || pkgname == NULL)
		return NULL;
	if (pi->comment[0] == '-')
		err += write_file(COMMENT_FNAME, &pi->comment[1]);
	else
		err += copy_file(PkgMgr.wrkdir, pi->comment, COMMENT_FNAME);
	if (pi->desc[0] == '-')
		err += write_file(DESC_FNAME, &pi->desc[1]);
	else
		err += copy_file(PkgMgr.wrkdir, pi->desc, DESC_FNAME);
	if (pi->install)
		err += copy_file(PkgMgr.wrkdir, pi->install, INSTALL_FNAME);
	if (pi->deinstall)
		err += copy_file(PkgMgr.wrkdir, pi->deinstall, DEINSTALL_FNAME);
	if (pi->require)
		err += copy_file(PkgMgr.wrkdir, pi->require, REQUIRE_FNAME);
	if (pi->display)
		err += copy_file(PkgMgr.wrkdir, pi->display, DISPLAY_FNAME);
	if (pi->mtree)
		err += copy_file(PkgMgr.wrkdir, pi->mtree, MTREE_FNAME);
	if (err > 0)
		return NULL;

	/*
	 * We fake making a contents file here.  It gets rewritten later, once
	 * it has gone through our normal processing in read_package().
	 */
	if ((pkg = new_package()) == NULL)
		return NULL;
	/* Turn the tarball name into a package name */
	if (pkgname_of(pkg->name, pkgname, PATH_MAX) == NULL) {
		free(pkg);
		return NULL;
	}
	if (pi->origin && strlen(pi->origin) > 0)
		strlcpy(pkg->origin, pi->origin, PATH_MAX);
	/* Copy the depends. */
	if (!PLIST_EMPTY(dl)) {
		msg_verbose("Registering depends:");
		PLIST_FOREACH(d, dl) {
			found = check_depend(d);
			add_depend_entry(&pkg->depends, d->type, d->file, d->pkg);
			msg_verbose(" %s%s", d->pkg, (found ? "" : " (not found)"));
		}
		msg_verbose(".\n");
	}
	/* Set the prefix.  If this is not set it is set to "." */
	if (pi->prefix && strlen(pi->prefix) > 0) {
		strlcpy(prefix, pi->prefix, PATH_MAX);
		add_plist_entry(&pkg->plist, PLIST_CWD, pi->prefix);
	} else
		strlcpy(prefix, ".", PATH_MAX);
	/* If a SrcDir override is set, add it now */
	if (pi->srcdir && strlen(pi->srcdir) > 0) {
		strlcpy(prefix, pi->srcdir, PATH_MAX);
		msg_verbose("Using source directory value of %s\n", pi->srcdir);
		add_plist_entry(&pkg->plist, PLIST_SRC, pi->srcdir);
	}
	/* Copy the packaging list, check it as we go */
	if (!strcmp(pi->contents, "-"))
		fp = stdin;
	else
		fp = fopen(pi->contents, "r");
	if (fp == NULL) {
		msg_warn("unable to open contents file '%s' for input", pi->contents);
		free_package(pkg);
		return NULL;
	}
	while (read_line(pline, PATH_MAX, fp) != NULL) {
		type = read_plist_entry(pline, &cp);
		if (type == PLIST_INVALID) {
			msg_warn("invalid packing list entry!");
			free_package(pkg);
			return NULL;
		}
		p = add_plist_entry(&pkg->plist, type, cp);
		if (p->type == PLIST_CWD || p->type == PLIST_SRC)
			snprintf(prefix, PATH_MAX, "%s", p->name);
		else if (p->type == PLIST_FILE) {
			snprintf(tmp, PATH_MAX, "%s/%s", prefix, p->name);
			if (!isnode(tmp))
				msg_warn("the file '%s' is not installed.", tmp);
			else
				p->md5 = make_md5(tmp);
		}
	}
	if (strcmp(pi->contents, "-"))
		fclose(fp);

	/* Now put the release specific items in */
	add_plist_entry(&pkg->plist, PLIST_CWD, ".");
	p = add_plist_entry(&pkg->plist, PLIST_FILE, COMMENT_FNAME);
	p->ignore = TRUE;
	p->md5 = make_md5(COMMENT_FNAME);
	p = add_plist_entry(&pkg->plist, PLIST_FILE, DESC_FNAME);
	p->ignore = TRUE;
	p->md5 = make_md5(DESC_FNAME);
	if (isnode(INSTALL_FNAME)) {
		p = add_plist_entry(&pkg->plist, PLIST_FILE, INSTALL_FNAME);
		p->ignore = TRUE;
		p->md5 = make_md5(INSTALL_FNAME);
	}
	if (isnode(DEINSTALL_FNAME)) {
		p = add_plist_entry(&pkg->plist, PLIST_FILE, DEINSTALL_FNAME);
		p->ignore = TRUE;
		p->md5 = make_md5(DEINSTALL_FNAME);
	}
	if (isnode(REQUIRE_FNAME)) {
		p = add_plist_entry(&pkg->plist, PLIST_FILE, REQUIRE_FNAME);
		p->ignore = TRUE;
		p->md5 = make_md5(REQUIRE_FNAME);
	}
	if (isnode(DISPLAY_FNAME)) {
		p = add_plist_entry(&pkg->plist, PLIST_FILE, DISPLAY_FNAME);
		p->ignore = TRUE;
		p->md5 = make_md5(DISPLAY_FNAME);
		strlcpy(pkg->display, DISPLAY_FNAME, PATH_MAX);
	}
	if (isnode(MTREE_FNAME)) {
		p = add_plist_entry(&pkg->plist, PLIST_FILE, MTREE_FNAME);
		p->ignore = TRUE;
		p->md5 = make_md5(MTREE_FNAME);
		strlcpy(pkg->mtree, MTREE_FNAME, PATH_MAX);
	}

	if (write_package(pkg) == FAIL) {
		msg_warn("failed to write package");
		free_package(pkg);
		return NULL;
	}
	return pkg;
}

/*
 * Get all of the info that we need from the package.  This just trashes
 * any existing pkg structure.
 *
 * c and n are the current and next pline.  We flip elements in the pline,
 * cp, and type arrays to avoid copies.  Yes, it's nasty, but it's
 * efficent...
 */
error_code
read_package(package *pkg)
{
	int c = 0, n = (c ? 0 : 1);
	char *cp[2], pline[2][PATH_MAX];
	plist_t type[2];
	boolean ignore = FALSE;
	FILE *fp;
	plist_entry *p;
	
	if (pkg == NULL)
		return FAIL;
	free_depend(&pkg->depends);
	free_require(&pkg->requires);
	free_plist(&pkg->plist);
	pkg->name[0] = '\0';
	pkg->origin[0] = '\0';
	pkg->display[0] = '\0';
	pkg->mtree[0] = '\0';
	pkg->options = PKG_OPTION_NONE;
	
	if ((fp = fopen(CONTENTS_FNAME, "r")) == NULL) {
		msg_warn("unable to open contents file!");
		return FAIL;
	}
	if (read_line(pline[c], PATH_MAX, fp) == NULL) {
		msg_warn("unable to read from contents file!");
		return FAIL;
	}
	while (strlen(pline[c]) > 0) {
		type[c] = read_plist_entry(pline[c], &cp[c]);
		read_line(pline[n], PATH_MAX, fp);
		switch(type[c]) {
		case PLIST_NAME:
			strlcpy(pkg->name, cp[c], PATH_MAX);
			break;
			
		case PLIST_ORIGIN:
			strlcpy(pkg->origin, cp[c], PATH_MAX);
			break;
			
		case PLIST_DISPLAY:
			strlcpy(pkg->display, cp[c], PATH_MAX);
			break;
			
		case PLIST_MTREE:
			strlcpy(pkg->mtree, cp[c], PATH_MAX);
			break;
			
		case PLIST_OPTION_INPLACE:
			pkg->options |= PKG_OPTION_INPLACE;
			break;
			
		case PLIST_OPTION_PRESERVE:
			pkg->options |= PKG_OPTION_PRESERVE;
			break;
			
		case PLIST_PKGDEP:
			type[n] = read_plist_entry(pline[n], &cp[n]);
			if (type[n] == PLIST_LIBDEP) {
				add_depend_entry(&pkg->depends, DEPEND_LIB, cp[n], cp[c]);
				read_line(pline[n], PATH_MAX, fp);
			} else if (type[n] == PLIST_RUNDEP) {
				add_depend_entry(&pkg->depends, DEPEND_RUN, cp[n], cp[c]);
				read_line(pline[n], PATH_MAX, fp);
			} else
				add_depend_entry(&pkg->depends, DEPEND_PKG, NULL, cp[c]);
			break;
		
		case PLIST_LIBDEP:
		case PLIST_RUNDEP:
			msg_warn("invalid dependency information");
			break;
			
		case PLIST_IGNORE:
		case PLIST_IGNORE_INST:
			ignore = TRUE;
			break;
			
		case PLIST_FILE:
			p = add_plist_entry(&pkg->plist, PLIST_FILE, cp[c]);
			if (ignore)
				p->ignore = TRUE;
			ignore = FALSE;
			type[n] = read_plist_entry(pline[n], &cp[n]);
			if (type[n] == PLIST_MD5) {
				p->md5 = copy_string(cp[n]);
				read_line(pline[n], PATH_MAX, fp);
			}
			break;

		case PLIST_CWD:
		case PLIST_SRC:
		case PLIST_CMD:
		case PLIST_UNEXEC:
		case PLIST_CHMOD:
		case PLIST_CHOWN:
		case PLIST_CHGRP:
		case PLIST_COMMENT:
		case PLIST_DIR_RM:
			add_plist_entry(&pkg->plist, type[c], cp[c]);
			break;
	
		case PLIST_INVALID:
		default:
			msg_warn("invalid packing list type");
			return FAIL;
		}
		c = (c ? 0 : 1), n = (c ? 0 : 1);
	}
	fclose(fp);

	if (!find_plist_entry(&pkg->plist, PLIST_CWD, NULL))
		add_plist_top(&pkg->plist, PLIST_CWD, ".");
	/* Protect against old packages with bogus @name fields */
	if (strlen(pkg->name) == 0)
		strlcpy(pkg->name, "anonymous", PATH_MAX);
	/*
	 * If we have a prefix override, delete the first one we see and add
	 * this one in place of it.
	 */
	if (strlen(PkgMgr.prefix_override) > 0) {
		p = find_plist_entry(&pkg->plist, PLIST_CWD, NULL);
		delete_plist_entry(&pkg->plist, p);
		add_plist_top(&pkg->plist, PLIST_CWD, PkgMgr.prefix_override);
	}
	if (!isemptyfile(REQUIRED_BY_FNAME)) {
		if ((fp = fopen(REQUIRED_BY_FNAME, "r")) == NULL) {
			msg_warn("can't open required_by file '%s'!", REQUIRED_BY_FNAME);
			return FAIL;
		}
		while (read_line(pline[0], PATH_MAX, fp))
			add_require_entry(&pkg->requires, pline[0]);
		fclose(fp);
	}
	return SUCCESS;
}

/*
 * Write out our structures.
 */
error_code
write_package(package *pkg)
{
	plist_entry *p;
	depend_entry *d;
	require_entry *r;
	FILE *fp;
	error_count err_cnt = 0;

	if (pkg == NULL || strlen(pkg->name) == 0)
		return FAIL;
	if (!(fp = fopen(CONTENTS_FNAME, "w"))) {
		msg_warn("can't open new contents file '%s'! can't register package", CONTENTS_FNAME);
		return FAIL;
	}
	err_cnt += write_plist_entry(fp, PLIST_NAME, pkg->name);
	if (strlen(pkg->origin) > 0)
		err_cnt += write_plist_entry(fp, PLIST_ORIGIN, pkg->origin);
	if (strlen(pkg->display) > 0)
		err_cnt += write_plist_entry(fp, PLIST_DISPLAY, pkg->display);
	if (strlen(pkg->mtree) > 0)
		err_cnt += write_plist_entry(fp, PLIST_MTREE, pkg->mtree);
	if (pkg->options & PKG_OPTION_INPLACE)
		err_cnt += write_plist_entry(fp, PLIST_OPTION_INPLACE, NULL);
	if (pkg->options & PKG_OPTION_PRESERVE)
		err_cnt += write_plist_entry(fp, PLIST_OPTION_PRESERVE, NULL);

	PLIST_FOREACH(d, &pkg->depends) {
		if (strlen(d->pkg) == 0)
			msg_warn("invalid pacakge dependency!");
		err_cnt += write_plist_entry(fp, PLIST_PKGDEP, d->pkg);
		switch (d->type) {
		case DEPEND_PKG:
			break;
		case DEPEND_LIB:
			err_cnt += write_plist_entry(fp, PLIST_LIBDEP, d->file);
			break;
		case DEPEND_RUN:
			err_cnt += write_plist_entry(fp, PLIST_RUNDEP, d->file);
			break;
		}
	}

	PLIST_FOREACH(p, &pkg->plist) {
		switch (p->type) {
		case PLIST_FILE:
			if (p->ignore)
				err_cnt += write_plist_entry(fp, PLIST_IGNORE, NULL);
			err_cnt += write_plist_entry(fp, p->type, p->name);
			if (p->md5)
				err_cnt += write_plist_entry(fp, PLIST_MD5, p->md5);
			break;

		case PLIST_CWD:
		case PLIST_SRC:
		case PLIST_CMD:
		case PLIST_UNEXEC:
		case PLIST_CHMOD:
		case PLIST_CHOWN:
		case PLIST_CHGRP:
		case PLIST_COMMENT:
		case PLIST_DIR_RM:
			err_cnt += write_plist_entry(fp, p->type, p->name);
			break;
	
		case PLIST_IGNORE:
		case PLIST_IGNORE_INST:
		case PLIST_NAME:
		case PLIST_ORIGIN:
		case PLIST_DISPLAY:
		case PLIST_MTREE:
		case PLIST_OPTION_INPLACE:
		case PLIST_OPTION_PRESERVE:
		case PLIST_PKGDEP:
		case PLIST_LIBDEP:
		case PLIST_RUNDEP:
			msg_warn("packing list contains nonsense items");
			break;
		
		default:
			msg_error(2, "%s: invalid packing list type.", __func__);
			return FAIL;
		}
	}
	fclose(fp);

	if (!PLIST_EMPTY(&pkg->requires)) {
		if (!(fp = fopen(REQUIRED_BY_FNAME, "w"))) {
			msg_warn("can't open required by file '%s'!", REQUIRED_BY_FNAME);
			return FAIL;
		}
		PLIST_FOREACH(r, &pkg->requires)
			fprintf(fp, "%s\n", r->pkg);
		fclose(fp);
	} else {
		if (isnode(REQUIRED_BY_FNAME))
			unlink(REQUIRED_BY_FNAME);
	}

	if (err_cnt != 0)
		return FAIL;
	return SUCCESS;
}

/*
 * Extract a package to it's final resting place.  This is just plain ugly. 
 * It works by building a list of files in where_args, and a list of files
 * to adjust permissions on in perm_args and then dumping this list through
 * a tar pipe when we need to change the output directory, or any of our
 * options.
 */

#define STARTSTRING "/usr/bin/tar cf - "
#define TOOBIG(str) ((strlen(str) + PATH_MAX + where_count > maxargs) \
					|| (strlen(str) + PATH_MAX + perm_count > maxargs))

#define PUSHOUT(todir) /* push out string */ \
		if (where_count > sizeof(STARTSTRING)-1) { \
			strcat(where_args, "| /usr/bin/tar --unlink -xf - -C "); \
			strcat(where_args, todir); \
			if (vsystem(where_args)) { \
				msg_error(2, "%s: can not invoke %d byte tar pipeline: %s", \
				  __func__, strlen(where_args), where_args); \
			} \
			strcpy(where_args, STARTSTRING); \
			where_count = sizeof(STARTSTRING)-1; \
		} \
		if (perm_count) { \
			apply_perms(todir, perm_args, m, o, g); \
			perm_args[0] = 0;\
			perm_count = 0; \
		}

error_code
extract_package(package *pkg)
{
	plist_entry *p;
	char *last_file, *dir;
	char *where_args, *perm_args;
	char *m, *o, *g;
	int maxargs, where_count = 0, perm_count = 0, add_count;
	char tmp[PATH_MAX];

	if (pkg == NULL)
		return FAIL;
	/* Just use half the argument space. */
	maxargs = sysconf(_SC_ARG_MAX) / 2;
	if ((where_args = alloca(maxargs)) == NULL) {
		msg_error(2, "%s: can't get argument list space", __func__);
		return FAIL;
	}
	if ((perm_args = alloca(maxargs)) == NULL) {
		msg_error(2, "%s: can't get argument list space", __func__);
		return FAIL;
	}
	strcpy(where_args, STARTSTRING);
	where_count = strlen(where_args);
	perm_args[0] = 0;

	/* Reset the world. */
	last_file = NULL;
	dir = ".";
	m = NULL;
	o = NULL;
	g = NULL;

	/* Do it. */
	PLIST_FOREACH(p, &pkg->plist) {
		switch(p->type) {
		case PLIST_FILE:
			if (p->ignore || p->merged)
				break;
			last_file = p->name;
			snprintf(tmp, PATH_MAX, "%s/%s", dir, p->name);
			msg_verbose("extract: %s\n", tmp);
			if (strrchr(p->name,'\'')) {
				msg_error(2, "%s: Bogus filename \"%s\"", __func__, p->name);
				return FAIL;
			}
			if (PkgMgr.fake)
				break;
			/* first try to rename it into place */
			if (isnode(tmp))
				/* XXX hack - if truly immutable, rename fails */
				chflags(tmp, 0);
			if (rename(p->name, tmp) == SUCCESS) {
				if (p->name[0] == '/' || TOOBIG(p->name))
					PUSHOUT(dir);
			} else {
				/* rename failed, try copying with a big tar command */
				if (p->name[0] == '/' || TOOBIG(p->name))
					PUSHOUT(dir);
				add_count = snprintf(&where_args[where_count],
				              maxargs - where_count, " '%s'", p->name);
				if (add_count > maxargs - where_count) {
					msg_error(2, "%s: oops, miscounted strings!", __func__);
					return FAIL;
				}
				where_count += add_count;
			}
			/* try to add to list of perms to be changed and run in bulk. */
			add_count = snprintf(&perm_args[perm_count],
			              maxargs - perm_count, "'%s' ", p->name);
			if (add_count > maxargs - perm_count) {
				msg_error(2, "%s: oops, miscounted strings!", __func__);
				return FAIL;
			}
			perm_count += add_count;
			break;

		case PLIST_CWD:
			msg_verbose("extract: CWD to %s\n", p->name);
			PUSHOUT(dir);
			dir = p->name;
			if (strcmp(dir, ".")) {
				if (!PkgMgr.fake && make_hierarchy(dir, m, o, g) == FAIL) {
					msg_error(2, "%s: unable to cwd to '%s'", __func__, dir);
					return FAIL;
				}
			}
			break;

		case PLIST_CMD:
			if ((strstr(p->name, "%B") || strstr(p->name, "%F") ||
				 strstr(p->name, "%f")) && last_file == NULL) {
				msg_error(2, "%s: no last file specified for '%s' command", __func__, p->name);
				return FAIL;
			}
			if (strstr(p->name, "%D") && dir == NULL) {
				msg_error(2, "%s: no directory specified for '%s' command", __func__, p->name);
				return FAIL;
			}
			format_cmd(tmp, PATH_MAX, p->name, dir, last_file);
			PUSHOUT(dir);
			msg_verbose("extract: execute '%s'\n", tmp);
			if (!PkgMgr.fake && vsystem(tmp))
				msg_warn("command '%s' failed", tmp);
			break;

		case PLIST_CHMOD:
			PUSHOUT(dir);
			m = p->name;
			break;

		case PLIST_CHOWN:
			PUSHOUT(dir);
			o = p->name;
			break;

		case PLIST_CHGRP:
			PUSHOUT(dir);
			g = p->name;
			break;

		default:
			break;
		}
	}
	PUSHOUT(dir);

	return SUCCESS;
}

/*
 * A quick and dirty hack so we don't have to deal with old style POST_
 * scripts.
 */
static error_code
wrap_legacy_scripts(char *script, char *post_script)
{
	char tmp[PATH_MAX];
	FILE *ftmp;
	int s;
	char *cp;
	
	if (script == NULL || post_script == NULL)
		return FAIL;
	snprintf(tmp, PATH_MAX, "%s.XXXXXX", script);
    if ((s = mkstemp(tmp)) == FAIL || (ftmp = fdopen(s, "w")) == NULL) {
		remove(tmp);
		msg_error(2, "couldn't open temp wrapper `%s'", tmp);
		return FAIL;
	}
	fprintf(ftmp, "#!/bin/sh\n\n");
	fprintf(ftmp, "case $2 in\nPRE-INSTALL)\nDEINSTALL)");
	if ((cp = read_file(script)) == NULL) {
		fclose(ftmp);
		remove(tmp);
		msg_error(2, "couldn't read script `%s'", script);
		return FAIL;
	}
	fprintf(ftmp, "%s;;", cp);
	free(cp);
	fprintf(ftmp, "POST-INSTALL)\nPOST-DEINSTALL)\n");
	if ((cp = read_file(post_script)) == NULL) {
		fclose(ftmp);
		remove(tmp);
		msg_error(2, "couldn't read script `%s'", post_script);
		return FAIL;
	}
	fprintf(ftmp, "%s;;", cp);
	free(cp);
	fprintf(ftmp, "esac\n");
	fclose(ftmp);
    if (rename(tmp, script) == FAIL) {
		msg_error(2, "error renaming `%s' to `%s'", tmp, script);
		return FAIL;
	}
    if (chmod(script, 0755) == FAIL) {
		msg_error(2, "couldn't change mode on script `%s'", script);
		return FAIL;
	}
    /* just in case */
	remove(post_script);
    remove(tmp);
	return SUCCESS;
}

/*
 * Installs the package into prefix.
 */
error_code
install_package(package *pkg)
{
	plist_entry *p;

	if (pkg == NULL)
		return FAIL;
	p = find_plist_entry(&pkg->plist, PLIST_CWD, NULL);
	setenv(PKG_PREFIX, p->name, 1);

	/* Look for the requirements file */
	if (isnode(REQUIRE_FNAME)) {
		vsystem("/bin/chmod +x %s", REQUIRE_FNAME);		/* be sure */
		msg_verbose("Running requirements file first for %s.\n", pkg->name);
		if (!PkgMgr.fake && vsystem("./%s %s INSTALL", REQUIRE_FNAME, pkg->name)) {
			msg_warn("package %s fails requirements %s", pkg->name,
				   PkgMgr.force ? "installing anyway" : "- not installed");
			if (!PkgMgr.force)
				return FAIL;
		}
	}

	if (isnode(POST_INSTALL_FNAME))
		wrap_legacy_scripts(INSTALL_FNAME, POST_INSTALL_FNAME);
	if (isnode(POST_DEINSTALL_FNAME))
		wrap_legacy_scripts(DEINSTALL_FNAME, POST_DEINSTALL_FNAME);

	/* If we're really installing, and have an installation file, run it */
	if (!PkgMgr.noinstall) {
		if (isnode(INSTALL_FNAME)) {
			vsystem("/bin/chmod +x %s", INSTALL_FNAME);		 /* make sure */
			msg_verbose("Running pre-install for %s..\n", pkg->name);
			if (!PkgMgr.fake && vsystem("./%s %s PRE-INSTALL", INSTALL_FNAME, pkg->name)) {
				msg_warn("install script returned error status");
				return FAIL;				/* nothing to uninstall yet */
			}
		}
	}

	/* Now finally extract the entire show if we're not going direct */
	if (!(pkg->options & PKG_OPTION_INPLACE) && !PkgMgr.fake)
		extract_package(pkg);

	if (!PkgMgr.fake && isnode(MTREE_FNAME)) {
		msg_verbose("Running mtree for %s..\n", pkg->name);
		p = find_plist_entry(&pkg->plist, PLIST_CWD, NULL);
		msg_verbose("mtree -U -f %s -d -e -p %s >%s\n", MTREE_FNAME, p ? p->name : "/", _PATH_DEVNULL);
		if (!PkgMgr.fake) {
			if (vsystem("/usr/sbin/mtree -U -f %s -d -e -p %s >%s", MTREE_FNAME, p ? p->name : "/", _PATH_DEVNULL))
				msg_warn("mtree returned a non-zero status - continuing");
		}
		unlink(MTREE_FNAME);
	}

	/* Run the installation script one last time? */
	if (!PkgMgr.noinstall && isnode(INSTALL_FNAME)) {
		vsystem("/bin/chmod +x %s", INSTALL_FNAME);		/* make sure */
		msg_verbose("Running post-install for %s..\n", pkg->name);
		if (!PkgMgr.fake && vsystem("./%s %s POST-INSTALL",
		                          INSTALL_FNAME, pkg->name)) {
			msg_warn("install script returned error status");
			return FAIL;
		}
		unlink(INSTALL_FNAME);
	}
	return SUCCESS;
}

/*
 * Register a package into the dbdir.
 */
error_code
register_package(package *pkg, char *dbdir)
{
	depend_entry *d;
	pkg_entry *pe;

	if (pkg == NULL || dbdir == NULL || !isdir(dbdir))
		return FAIL;
	move_file(".", DESC_FNAME, dbdir);
	move_file(".", COMMENT_FNAME, dbdir);
	if (isnode(DEINSTALL_FNAME))
		move_file(".", DEINSTALL_FNAME, dbdir);
	if (isnode(REQUIRE_FNAME))
		move_file(".", REQUIRE_FNAME, dbdir);
	if (isnode(DISPLAY_FNAME))
		move_file(".", DISPLAY_FNAME, dbdir);

	if (enter_dir(dbdir) == FAIL) {
		msg_warn("couldn't enter the dbdir!");
		return FAIL;
	}
	if (write_package(pkg) == FAIL) {
		msg_warn("couldn't write packcage!");
		return FAIL;
	}
	PLIST_FOREACH(d, &pkg->depends) {
		if (strlen(d->pkg) == 0) {
			msg_warn("invalid package dependency");
			return FAIL;
		}
		msg_verbose("Attempting to record dependency on package `%s'\n", d->pkg);
		if ((pe = find_pkg_entry(d->pkg)) == NULL) {
			msg_warn("unable to find package '%s'", d->pkg);
			return FAIL;
		}
		if (scan_pkg_entry(pe) == FAIL)
			return FAIL;
		add_require_entry(&pe->pkg->requires, pkg->name);
		pe->status = PKG_CHANGED;
		if (write_pkg_entry(pe) == FAIL)
			return FAIL;
		cleanup_pkg_entry(pe);
	}
	return SUCCESS;
}

/*
 * Check the package's dependencies, and install them as needed.
 */
error_code
depend_package(package *pkg, char *hint)
{
	char *cp = NULL;
	int err = 0;
	depend_entry *d;
	boolean found = FALSE;

	if (pkg == NULL)
		return FAIL;
	PLIST_FOREACH(d, &pkg->depends) {
		found = check_depend(d);
		switch(d->type) {
		case DEPEND_PKG:
			asprintf(&cp, "package `%s'", d->pkg);
			break;

		case DEPEND_LIB:
			asprintf(&cp, "shared library %s supplied by `%s'", d->file, d->pkg);
			break;

		case DEPEND_RUN:
			asprintf(&cp, "executable %s supplied by `%s'", d->file, d->pkg);
			break;
		}
		if (cp != NULL) {
			msg_verbose("Package `%s' depends on %s.\n", pkg->name, cp);
			free(cp);
		}
		if (!found) {
			if (!PkgMgr.fake) {
				if ((cp = find_file(hint, d->pkg)) != NULL) {
					msg_verbose("Loading it from %s.\n", cp);
					if (pkg_mgr_add(cp)) {
						msg_warn("autoload of dependency `%s' failed%s",
								d->pkg, PkgMgr.force ? " (proceeding anyway)" : "!");
						if (!PkgMgr.force)
							err++;
					} else
						msg_verbose("\t`%s' loaded successfully.\n", d->pkg);
					free(cp);
				} else {
					msg_warn("could not find package %s %s",
							d->pkg, PkgMgr.force ? " (proceeding anyway)" : "!");
					if (!PkgMgr.force)
						err++;
				}
			} else {
				if (PkgMgr.verbose)
					msg_verbose("and was not found%s.\n", 
						   PkgMgr.force ? " (proceeding anyway)" : "");
				else
					printf("Package dependency %s not found%s\n", d->pkg,
						   PkgMgr.force ? " (proceeding anyway)" : "!");
				if (!PkgMgr.force)
					err++;
			}
		} else
			msg_verbose(" - already installed.\n");
	}

	if (err != 0)
		return FAIL;
	return SUCCESS;
}

/*
 * Remove pkg's dependencies on other packages.
 */
error_code
undepend_package(package *pkg)
{
	depend_entry *d;
	pkg_entry *pe;

	if (pkg == NULL)
		return FAIL;
	PLIST_FOREACH(d, &pkg->depends) {
		msg_verbose("Attempting to remove dependency on package `%s'\n", d->pkg);
		if (!PkgMgr.fake && check_depend(d)) {
			if ((pe = find_pkg_entry(d->pkg)) == NULL) {
				msg_warn("couldn't open package `%s'", d->pkg);
				continue;
			}
			if (scan_pkg_entry(pe)) {
				msg_warn("couldn't read package `%s'", d->pkg);
				return FAIL;
			}
			delete_require_entry(&pe->pkg->requires, pkg->name);
			pe->status = PKG_CHANGED;
			if (write_pkg_entry(pe))
				return FAIL;
			cleanup_pkg_entry(pe);
		}
	}	
	return SUCCESS;
}	

/*
 * Upgrade one package with another.  This just involves moving files from
 * one list to the other.  The deletion and addition are taken care of
 * in upgrade_pkg_entry().
 */
error_code
upgrade_package(package *from, package *to)
{
	plist_entry *p;
	
	if (from == NULL || to == NULL)
		return FAIL;
	PLIST_FOREACH(p, &from->plist) {
		if (p->type == PLIST_FILE && p->name && strstr(p->name, ".so.") != NULL) {
			if (move_plist_entry(&from->plist, p, &to->plist))
				return FAIL;
			p->merged = TRUE;
		}
	}
	return SUCCESS;
}

/*
 * Checks is a package is required.
 */
boolean
required_package(package *pkg)
{
	require_entry *r;

	if (pkg == NULL)
		return FALSE;
	if (!PLIST_EMPTY(&pkg->requires)) {
		msg_warn("package `%s' is required by these other packages\n"
		      "and may not be deinstalled%s:",
				pkg->name, PkgMgr.force ? " (but I'll delete it anyway)" : "" );
		PLIST_FOREACH(r, &pkg->requires)
			printf("%s\n", r->pkg);
		return TRUE;
	} else
		return FALSE;
}

/*
 * Delete the results of a package installation.
 */
error_code
delete_package(package *pkg)
{
	plist_entry *p;
	char *where = ".", *last_file = "";
	error_code fail = SUCCESS;
	char tmp[PATH_MAX];

	if (pkg == NULL)
		return FAIL;
	if ((p = find_plist_entry(&pkg->plist, PLIST_CWD, NULL)) == NULL) {
		msg_warn("package '%s' doesn't have a prefix", pkg->name);
		return FAIL;
	}
	setenv(PKG_PREFIX, p->name, 1);

	if (isnode(REQUIRE_FNAME)) {
		msg_verbose("Executing 'require' script.\n");
		vsystem("/bin/chmod +x %s", REQUIRE_FNAME);		/* be sure */
		if (vsystem("./%s %s DEINSTALL", REQUIRE_FNAME, pkg->name)) {
			msg_warn("package %s fails requirements %s", pkg->name,
				   PkgMgr.force ? "" : "- not deleted");
			if (!PkgMgr.force)
				return TRUE;
		}
	}

	if (isnode(POST_DEINSTALL_FNAME))
		wrap_legacy_scripts(DEINSTALL_FNAME, POST_DEINSTALL_FNAME);

	if (!PkgMgr.noinstall && isnode(DEINSTALL_FNAME)) {
		if (PkgMgr.fake)
			printf("Would execute de-install script at this point.\n");
		else {
			vsystem("/bin/chmod +x %s", DEINSTALL_FNAME);		/* make sure */
			if (vsystem("./%s %s DEINSTALL", DEINSTALL_FNAME, pkg->name)) {
				msg_warn("deinstall script returned error status");
				if (!PkgMgr.force)
					return TRUE;
			}
		}
	}

	PLIST_FOREACH(p, &pkg->plist) {
		switch (p->type)  {
		case PLIST_CWD:
			where = p->name;
			msg_verbose("Change working directory to %s\n", where);
			break;

		case PLIST_UNEXEC:
			format_cmd(tmp, PATH_MAX, p->name, where, last_file);
			msg_verbose("Execute `%s'\n", tmp);
			if (!PkgMgr.fake && vsystem(tmp)) {
				msg_warn("unexec command for `%s' failed", tmp);
				fail = FAIL;
			}
			break;

		case PLIST_FILE:
			if (p->ignore)
				break;
			last_file = p->name;
			snprintf(tmp, PATH_MAX, "%s/%s", where, p->name);
			if (isnode(tmp) && isdir(tmp) && !issymlink(tmp)) {	
				msg_warn("cannot delete specified file `%s' - it is a directory!\n"
			      "this packing list is incorrect - ignoring delete request", tmp);
				break;
			}
			if (p->md5 && !issymlink(tmp) && !check_md5(tmp, p->md5)) {
				printf("%s fails original MD5 checksum - %s\n", tmp, PkgMgr.force ? "deleted anyway." : "not deleted.");
  				if (!PkgMgr.force) {
					fail = FAIL;
					break;
				}
			}
			msg_verbose("Deleting file %s\n", tmp);
			if (!PkgMgr.fake) {
				if (delete_hierarchy(tmp, PkgMgr.force, PkgMgr.cleandirs))
					fail = FAIL;
			}
			break;

		case PLIST_DIR_RM:
			snprintf(tmp, PATH_MAX, "%s/%s", where, p->name);
			if (isnode(tmp) && !isdir(tmp)) {
				msg_warn("cannot delete specified directory `%s' - it is a file!\n"
				  "this packing list is incorrect - ignoring delete request", tmp);
			} else {
				msg_verbose("Delete directory %s\n", tmp);
				if (!PkgMgr.fake && delete_hierarchy(tmp, PkgMgr.force, FALSE)) {
					msg_warn("unable to completely remove directory '%s'", tmp);
					fail = FAIL;
				}
			}
			last_file = p->name;
			break;

		default:
			break;
		}
	}
	
	if (fail == FAIL) {
		msg_warn("couldn't entirely delete package (perhaps the packing"
		      "list is incorrectly specified?)");
	}
	
	if (!PkgMgr.noinstall && isnode(DEINSTALL_FNAME)) {
		if (PkgMgr.fake)
			printf("Would execute post-deinstall script at this point.\n");
		else {
			vsystem("/bin/chmod +x %s", DEINSTALL_FNAME);		/* make sure */
			if (vsystem("./%s %s POST-DEINSTALL", DEINSTALL_FNAME, pkg->name)) {
				msg_warn("post-deinstall script returned error status");
				if (!PkgMgr.force)
					return TRUE;
			}
		}
	}
	if (undepend_package(pkg) == FAIL)
		return FAIL;
	return SUCCESS;
}

/*
 * Make a package from installed files.
 */
error_code
make_package(package *pkg, char *pkgname)
{
	plist_entry *p;
	int ret;
	char *args[50];		/* Much more than enough. */
	int nargs = 0;
	int pipefds[2];
	FILE *totar;
	pid_t pid;

	if (pkg == NULL || pkgname == NULL)
		return FAIL;
	args[nargs++] = "tar";		/* argv[0] */
	args[nargs++] = "-c";
	args[nargs++] = "-f";
	args[nargs++] = pkgname;
	if (strchr(&pkgname[strlen(pkgname)-1], 'z'))		/* Compress/gzip? */
		args[nargs++] = "-z";
	else if (strchr(&pkgname[strlen(pkgname)-1], '2'))		/* Compress/bzip? */
		args[nargs++] = "-y";
	if (PkgMgr.dereference)
		args[nargs++] = "-h";
	if (PkgMgr.excludefrom) {
		args[nargs++] = "-X";
		args[nargs++] = PkgMgr.excludefrom;
	}
	args[nargs++] = "-T";		/* Take filenames from file instead of args. */
	args[nargs++] = "-";		/* Use stdin for the file. */
	args[nargs] = NULL;

	if (write_package(pkg))
		return FAIL;
	msg_verbose("Creating package as '%s'\n", pkgname);

	/* Set up a pipe for passing the filenames, and fork off a tar process. */
	if (pipe(pipefds) == -1) {
		msg_error(2, "%s: cannot create pipe", __func__);
		return FAIL;
	}
	if ((pid = fork()) == -1) {
		msg_error(2, "%s: cannot fork process for tar", __func__);
		return FAIL;
	}
	if (pid == 0) {		/* The child */
		dup2(pipefds[0], 0);
		close(pipefds[0]);
		close(pipefds[1]);
		execv("/usr/bin/tar", args);
		msg_error(2, "%s: failed to execute tar command", __func__);
		return FAIL;
	}

	/* Meanwhile, back in the parent process ... */
	close(pipefds[0]);
	if ((totar = fdopen(pipefds[1], "w")) == NULL) {
		msg_error(2, "%s: fdopen failed", __func__);
		return FAIL;
	}
    fprintf(totar, "%s\n", CONTENTS_FNAME);
    fprintf(totar, "%s\n", COMMENT_FNAME);
    fprintf(totar, "%s\n", DESC_FNAME);
    if (isnode(INSTALL_FNAME))
		fprintf(totar, "%s\n", INSTALL_FNAME);
    if (isnode(DEINSTALL_FNAME))
		fprintf(totar, "%s\n", DEINSTALL_FNAME);
    if (isnode(REQUIRE_FNAME))
		fprintf(totar, "%s\n", REQUIRE_FNAME);
    if (isnode(DISPLAY_FNAME))
		fprintf(totar, "%s\n", DISPLAY_FNAME);
    if (isnode(MTREE_FNAME))
		fprintf(totar, "%s\n", MTREE_FNAME);
	PLIST_FOREACH(p, &pkg->plist) {
		if (p->type == PLIST_FILE && !p->ignore)
			fprintf(totar, "%s\n", p->name);
		else if (p->type == PLIST_CWD || p->type == PLIST_SRC) 
			fprintf(totar, "-C\n%s\n", p->name);
	}
	fclose(totar);
	wait(&ret);
	/* assume either signal or bad exit is enough for us */
	if (ret) {
		msg_error(2, "%s: tar command failed with code %d", __func__, ret);
		return FAIL;
	}
	return SUCCESS;
}

/*
 * Output pretty information about a package.  XXX: (reg) This should be in
 * show.c.
 */
error_code
show_package(package *pkg, unsigned int flags)
{
	error_count err_cnt = 0;

	if (pkg == NULL)
		return FAIL;
	if (flags & PKG_SHOW_INDEX || flags & PKG_SHOW_WHICH) {
		msg_warn("invalid show types for package");
		return TRUE;
	}

	/* Start showing the package contents */
	msg_verbose("%sInformation for %s:\n\n", PkgMgr.infoprefix, pkg->name);
	if (flags & PKG_SHOW_COMMENT)
		err_cnt += show_file("Comment:\n", COMMENT_FNAME);
	if (flags & PKG_SHOW_REQUIRE)
		err_cnt += show_depend("Depends on:\n", &pkg->depends);
	if ((flags & PKG_SHOW_REQBY) && !isemptyfile(REQUIRED_BY_FNAME))
		err_cnt += show_file("Required by:\n", REQUIRED_BY_FNAME);
	if (flags & PKG_SHOW_DESC)
		err_cnt += show_file("Description:\n", DESC_FNAME);
	if ((flags & PKG_SHOW_DISPLAY) && isnode(DISPLAY_FNAME))
		err_cnt += show_file("Install notice:\n", DISPLAY_FNAME);
	if (flags & PKG_SHOW_PLIST)
		err_cnt += show_plist("Packing list:\n", pkg, (plist_t)-1);
	if ((flags & PKG_SHOW_INSTALL) && isnode(INSTALL_FNAME))
		err_cnt += show_file("Install script:\n", INSTALL_FNAME);
	if ((flags & PKG_SHOW_INSTALL) && isnode(POST_INSTALL_FNAME))
		err_cnt += show_file("Post-Install script:\n", POST_INSTALL_FNAME);
	if ((flags & PKG_SHOW_DEINSTALL) && isnode(DEINSTALL_FNAME))
		err_cnt += show_file("De-Install script:\n", DEINSTALL_FNAME);
	if ((flags & PKG_SHOW_DEINSTALL) && isnode(POST_DEINSTALL_FNAME))
		err_cnt += show_file("Post-DeInstall script:\n", POST_DEINSTALL_FNAME);
	if ((flags & PKG_SHOW_MTREE) && isnode(MTREE_FNAME))
		err_cnt += show_file("mtree file:\n", MTREE_FNAME);
	if (flags & PKG_SHOW_PREFIX)
		err_cnt += show_plist("Prefix(s):\n", pkg, PLIST_CWD);
	if (flags & PKG_SHOW_FILES)
		err_cnt += show_files("Files:\n", pkg);
	if ((flags & PKG_SHOW_SIZE))
		err_cnt += show_size("Package Size:\n", pkg);
	if ((flags & PKG_SHOW_CHKSUM))
		err_cnt += show_chksum("Package Checksums:\n", pkg);
	if (flags & PKG_SHOW_ORIGIN)
		err_cnt += show_origin("Origin:\n", pkg);
	msg_verbose(PkgMgr.infoprefix);
	
	if (err_cnt != 0)
		return FAIL;
	return FALSE;
}

/*
 * Try to determine the version of a package compared to the latest version. 
 * We take the package name we are passed and first slurp in the CONTENTS
 * file, getting name and origin, then we look for it's corresponding
 * Makefile. If that fails we pull in the INDEX, and check there.
 */
error_code
version_package(package *pkg)
{
	char tmp[PATH_MAX], tmp2[PATH_MAX], *latest = NULL;
	index_entry *ie;

	/*
	 * First we check if the installed package has an origin, and try
	 * looking for it's Makefile. If we find the Makefile we get the
	 * latest version from there. If we fail, we start looking in the
	 * INDEX, first matching the origin and then the package name.
	 */
	if (strlen(pkg->origin) > 0) {
		snprintf(tmp, PATH_MAX, "%s/%s", PORTS_DIR, pkg->origin);
		if (isdir(tmp) && chdir(tmp) != FAIL && isfile("Makefile")) {
			if ((latest = vpipe("make -V PKGNAME", tmp)) == NULL)
				warnx("Failed to get PKGNAME from %s/Makefile!", tmp);
			else
				show_version(pkg->name, latest, "port");
		}
	}
	if (latest == NULL) {
		/* We only pull in the INDEX once, if needed. */
		if (PLIST_EMPTY(&PkgIndex))
			pkg_mgr_index();
		/* Now that we've slurped in the INDEX... */
		PLIST_FOREACH(ie, &PkgIndex) {
			if (strlen(pkg->origin) > 0) {
				if (strcmp(pkg->origin, ie->origin) == 0)
					latest = copy_string(ie->name);
			} else {
				/* Chop off the versions and compare. */
				if (pkgroot_of(tmp, ie->name, PATH_MAX) == NULL)
					msg_error(2, "The INDEX does not appear to be valid!");
				if (pkgroot_of(tmp2, pkg->name, PATH_MAX) == NULL)
					msg_warn("%s is not a valid package!", pkg->name);
				if (strcmp(tmp, tmp2) == 0) {
					if (latest != NULL) {
						/* Multiple matches */
						snprintf(tmp, PATH_MAX, "%s|%s", latest, ie->name);
						free(latest);
						latest = copy_string(tmp);
					} else
						latest = copy_string(ie->name);
				}
			}
		}
		if (latest == NULL)
			show_version(pkg->name, NULL, pkg->origin);
		else
			show_version(pkg->name, latest, "index");
	}
	if (latest != NULL)
		free(latest);
	return SUCCESS;
}
