/*
 * XXX: (reg) insert final project name here.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Jeremy D. Lea
 * XXX: (reg) Insert date
 *
 * Based on pkg_install, by:
 *   Jordan K. Hubbard
 *   18 July 1993
 *
 * General packing list routines.
 */

#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <md5.h>
#include "list.h"
#include "lib.h"

/*
 * Allocate a new packing list entry.
 */
plist_entry *
new_plist_entry(void)
{
	plist_entry *p;

	if ((p = (plist_entry *)malloc(sizeof(plist_entry))) == NULL) {
		msg_error(2, "%s: unable to allocate memory", __func__);
		return NULL;
	}	
	bzero(p, sizeof(plist_entry));
	return p;
}

/*
 * Free an entire packing list.
 */
void
free_plist(plist *pl)
{
	plist_entry *tmp;

	if (pl == NULL)
		return;
	while (!PLIST_EMPTY(pl)) {
		PLIST_REMOVE_HEAD(pl, tmp);
		if (tmp->name !=NULL)
			free(tmp->name);
		if (tmp->md5 != NULL)
			free(tmp->md5);
		free(tmp);
	}
}

/*
 * Find a given item in a packing list and, if so, return it (else NULL).
 */
plist_entry *
find_plist_entry(plist *pl, plist_t type, char *name)
{
	plist_entry *p;

	if (pl == NULL)
		return NULL;
	PLIST_FOREACH(p, pl) {
		if (p->type == type) {
			if (name != NULL) {
				if (!strcmp(p->name, name))
	    			return p;
			} else
				return p;
		}
	}
	return NULL;
}

/*
 * Add an item to a packing list.
 */
plist_entry *
add_plist_entry(plist *pl, plist_t type, char *arg)
{
	plist_entry *p;
	
	if (pl == NULL)
		return NULL;
	if ((p = new_plist_entry()) == NULL)
		return NULL;
	p->name = copy_string(arg);
	p->type = type;
	PLIST_INSERT_TAIL(pl, p);
	return p;
}

/*
 * Add an item to the top of a packing list.
 */
plist_entry *
add_plist_top(plist *pl, plist_t type, char *arg)
{
	plist_entry *p;

	if (pl == NULL)
		return NULL;
	if ((p = new_plist_entry()) == NULL)
		return NULL;
	p->name = copy_string(arg);
	p->type = type;
	PLIST_INSERT_HEAD(pl, p);
	return p;
}

/*
 * Delete an entry from the plist.
 */
error_code
delete_plist_entry(plist *pl, plist_entry *p)
{
	plist_entry *tmp = NULL;
	
	if (pl == NULL || p == NULL)
		return FAIL;
	PLIST_REMOVE(pl, p, tmp);
	if (p == NULL)
		return FAIL;
	if (p->name != NULL)
		free(p->name);
	if (p->md5 != NULL)
		free(p->md5);
	free(p);
	return SUCCESS;
}

/*
 * Read a packing list entry from a string
 */
plist_t
read_plist_entry(char *pline, char **name)
{
	char *cp, *sp;
	int len;
   	char cmd[PATH_MAX];

   	if (pline == NULL || name == NULL) {
		msg_error(2, "%s: null plist entry or name buffer", __func__);
		return PLIST_INVALID;
	}
	len = strlen(pline);
	while (len > 0 && isspace(pline[len-1]))
		pline[--len] = '\0';
	if (len == 0)
		return PLIST_INVALID;
	cp = pline;
	sp = cmd;
	if (*cp == CMD_CHAR) {
    	while (*++cp && isalnum(*cp))
			*sp++ = *cp;
		*sp = '\0';
    	while (*++cp && isspace(*cp))
			;
		if (*cp == '\0')
			*name = NULL;
		else
			*name = cp;
    	if (!strcmp(cmd, "cwd"))
    		return PLIST_CWD;
    	else if (!strcmp(cmd, "srcdir"))
    		return PLIST_SRC;
    	else if (!strcmp(cmd, "cd"))
    		return PLIST_CWD;
    	else if (!strcmp(cmd, "exec"))
    		return PLIST_CMD;
    	else if (!strcmp(cmd, "unexec"))
    		return PLIST_UNEXEC;
    	else if (!strcmp(cmd, "mode"))
    		return PLIST_CHMOD;
    	else if (!strcmp(cmd, "owner"))
    		return PLIST_CHOWN;
    	else if (!strcmp(cmd, "group"))
    		return PLIST_CHGRP;
    	else if (!strcmp(cmd, "ignore"))
    		return PLIST_IGNORE;
    	else if (!strcmp(cmd, "ignore_inst"))
    		return PLIST_IGNORE_INST;
    	else if (!strcmp(cmd, "name"))
    		return PLIST_NAME;
    	else if (!strcmp(cmd, "display"))
    		return PLIST_DISPLAY;
    	else if (!strcmp(cmd, "pkgdep"))
    		return PLIST_PKGDEP;
    	else if (!strcmp(cmd, "mtree"))
    		return PLIST_MTREE;
    	else if (!strcmp(cmd, "dirrm"))
    		return PLIST_DIR_RM;
    	else if (!strcmp(cmd, "option")) {
    		if (cp && !strcmp(cp, "extract-in-place"))
    			return PLIST_OPTION_INPLACE;
    		else if (cp && !strcmp(cp, "preserve"))
    			return PLIST_OPTION_PRESERVE;
    		else
    			return PLIST_INVALID;
			}
    	else if (!strcmp(cmd, "comment")) {
			if (!cp)
				return PLIST_COMMENT;
			/* Extended comment support. */
			sp = strchr(cp, ':'); 
			if (sp++ && *sp != '\0') {
				*name = sp;
				if (!strncmp(cp, "MD5:", sp - cp))
					return PLIST_MD5;
				else if (!strncmp(cp, "ORIGIN:", sp - cp))
					return PLIST_ORIGIN;
				else if (!strncmp(cp, "LIBDEP:", sp - cp))
					return PLIST_LIBDEP;
				else if (!strncmp(cp, "RUNDEP:", sp - cp))
					return PLIST_RUNDEP;
				else
					*name = cp;
    		}
    		return PLIST_COMMENT;
		}
    	else {
			msg_warn("bad command '%s'", pline);
    		return PLIST_INVALID;
		}
	}
	else {
		*name = pline;
		return PLIST_FILE;
	}
}

/*
 * Write a packing list to a file, converting commands to ascii equivalents.
 */
error_code
write_plist_entry(FILE *fp, plist_t type, char *name)
{

	if (fp == NULL)
		return FAIL;
	switch (type) {
	case PLIST_FILE:
		fprintf(fp, "%s", name);
		break;

	case PLIST_CWD:
		fprintf(fp, "%ccwd %s", CMD_CHAR, name);
		break;

	case PLIST_SRC:
		fprintf(fp, "%csrcdir %s", CMD_CHAR, name);
		break;

	case PLIST_CMD:
		fprintf(fp, "%cexec %s", CMD_CHAR, name);
		break;

	case PLIST_UNEXEC:
		fprintf(fp, "%cunexec %s", CMD_CHAR, name);
		break;

	case PLIST_CHMOD:
		fprintf(fp, "%cmode %s", CMD_CHAR, name ? name : "");
		break;

	case PLIST_CHOWN:
		fprintf(fp, "%cowner %s", CMD_CHAR, name ? name : "");
		break;

	case PLIST_CHGRP:
		fprintf(fp, "%cgroup %s", CMD_CHAR, name ? name : "");
		break;

	case PLIST_IGNORE:
	case PLIST_IGNORE_INST:				/* a one-time non-ignored file */
		fprintf(fp, "%cignore", CMD_CHAR);
		break;

	case PLIST_PKGDEP:
		fprintf(fp, "%cpkgdep %s", CMD_CHAR, name);
		break;

	case PLIST_DIR_RM:
		fprintf(fp, "%cdirrm %s", CMD_CHAR, name);
		break;

	case PLIST_NAME:
		fprintf(fp, "%cname %s", CMD_CHAR, name);
		break;

	case PLIST_DISPLAY:
		fprintf(fp, "%cdisplay %s", CMD_CHAR, name);
		break;

	case PLIST_MTREE:
		fprintf(fp, "%cmtree %s", CMD_CHAR, name);
		break;

	case PLIST_OPTION_INPLACE:
		fprintf(fp, "%coption extract-in-place", CMD_CHAR);
		break;

	case PLIST_OPTION_PRESERVE:
		fprintf(fp, "%coption preserve", CMD_CHAR);
		break;

	case PLIST_COMMENT:
		fprintf(fp, "%ccomment %s", CMD_CHAR, name);
		break;

	/* Extended comment support. */
	case PLIST_MD5:
		fprintf(fp, "%ccomment MD5:%s", CMD_CHAR, name);
		break;
		
	case PLIST_ORIGIN:
		fprintf(fp, "%ccomment ORIGIN:%s", CMD_CHAR, name);
		break;

	case PLIST_LIBDEP:
		fprintf(fp, "%ccomment LIBDEP:%s", CMD_CHAR, name);
		break;

	case PLIST_RUNDEP:
		fprintf(fp, "%ccomment RUNDEP:%s", CMD_CHAR, name);
		break;

	case PLIST_INVALID:
	default:
		msg_error(2, "unknown command type %d (%s)", type, name ? name : "");
		return FAIL;
	}
	fprintf(fp, "\n");
	return SUCCESS;
}

/*
 * Move a plist entry from one plist to another, trying to keep the list
 * ordered.
 */
error_code
move_plist_entry(plist *from, plist_entry *move, plist *to)
{
	int i;
	plist_entry *p, *l = NULL;
	
	PLIST_REMOVE(from, move, l);
	/* l will be NULL if we didn't find the element 'move' in 'from'. */
	if (l == NULL)
		return FAIL;
	/*
	 * We try to insert after the first entry that is the same as the
	 * element we were after in the last list, but we don't insert if there
	 * is an existing entry which has the same name.
	 */
	PLIST_FOREACH(p, to) {
		if (l->name != NULL && p->name != NULL) {
			if ((i = strcmp(l->name, p->name)) == 0)
				return SUCCESS;
			else if (i > 0) {
				PLIST_INSERT_AFTER(to, p, move);
				return SUCCESS;
			}
		}
	}
	PLIST_INSERT_TAIL(to, move);
	return SUCCESS;
}
