/*
 * XXX: (reg) insert final project name here.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Jeremy D. Lea
 * XXX: (reg) Insert date
 *
 * Based on pkg_install, by:
 *   Jordan K. Hubbard
 *   18 July 1993
 *
 * This file contains code for handling strings (generally package names).
 */

#include <sys/param.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "list.h"
#include "lib.h"

/*
 * Return the filename portion of a path.
 */
const char *
basename_of(const char *str)
{
	int i;

	if (str == NULL)
		return NULL;
	i = strlen(str);
	while (i > 0 && str[i-1] != '/')
		i--;
	if (strlen(&str[i]) == 0)
		return NULL;
	return &str[i];
}

/*
 * Clean up pkgname and return it in str.
 */
char *
pkgname_of(char *str, const char *pkgname, int max)
{
	int i;
	
	if (pkgname == NULL || (i = strlen(pkgname)) == 0)
		return NULL;
	/* Start from the end looking for [/][A-Za-z0-9]. */
	while (--i > 0 && (pkgname[i-1] != '/' || !isalnum(pkgname[i])))
		;
	/* Go forward until we have [A-Za-z0-9]. */
	while (pkgname[i] != '\0' && !isalnum(pkgname[i]))
		i++;
	strlcpy(str, &pkgname[i], max);
	if ((i = strlen(str)) == 0)
		return NULL;
	while (--i >= 0) {
		/* We're very selective about the characters we associate with. */
		if (!isalnum(str[i]) && strchr("+,-._", str[i]) == NULL)
			str[i] = '\0';
		/* We chop off anything that looks like a file extention. */
		/* XXX: (reg) bzip packages too... */
		if (str[i] == '.' && (!strcmp(&str[i], PKG_SUFFIX)
		 || !strcmp(&str[i], ".tgz") || !strcmp(&str[i], ".tar")
		 || !strcmp(&str[i], ".Z") || !strcmp(&str[i], ".gz")))
			str[i] = '\0';
		/* Now kill anything not [A-Za-z0-9] at the end. */
		if (str[i+1] == '\0' && !isalnum(str[i]))
			str[i] = '\0';
	}
	if (strlen(str) == 0)
		return NULL;
	return str;
}

/*
 * Work out the root of the pkgname and return it in str.
 */
char *
pkgroot_of(char *str, const char *pkgname, int max)
{
	char *cp;
	
	if (pkgname_of(str, pkgname, max) == NULL)
		return NULL;
	if ((cp = strrchr(str, '-')) != NULL)
		cp[0] = '\0';
	if (strlen(str) == 0)
		return NULL;
	return str;
}

/*
 * Returns a pointer to the version portion of a package name and the two
 * special components.  If the version does not appear to be valid, we
 * return NULL, and the epoch and revision are undefined.
 */
const char *
version_of(const char *pkgname, int *epoch, int *revision)
{
	int i, e = -1, r = -1;

	if (pkgname == NULL || (i = strlen(pkgname)) == 0)
		return NULL;
	while (--i >= 0 && pkgname[i] != '-') {
		/* Check if we really have a version number. */
		if (!isalnum(pkgname[i]) && strchr("_,.", pkgname[i]) == NULL)
			return NULL;
		if (pkgname[i] == ',') {
			if (r != -1 || e != -1)
				return NULL;
			e = atoi(&pkgname[i+1]);
		}
		if (pkgname[i] == '_') {
			if (r != -1)
				return NULL;
			r = atoi(&pkgname[i+1]);
		}
	}
	if (strlen(&pkgname[++i]) == 0)
		return NULL;
	if (epoch != NULL)
		*epoch = (e == -1 ? 0 : e);
	if (revision != NULL)
		*revision = (r == -1 ? 0 : r);
	return &pkgname[i];
}

/*
 * Returns -1, 0 or 1 depending on if the version components of pkg1 is less
 * than, equal to or greater than pkg2. No comparision of the basenames is
 * done.
 *
 * The port verison is defined by:
 * ${PORTVERSION}[_${PORTREVISION}][,${PORTEPOCH}] {PORTEPOCH} supercedes
 * ${PORTVERSION} supercedes ${PORTREVISION}. See the commit log for
 * revision 1.349 of ports/Mk/bsd.port.mk for more information.
 *
 * The epoch and revision are defined to be a single number, while the rest
 * of the version should conform to the porting guidelines. It can contain
 * multiple components, seperated by a period, including letters.
 *
 * The tests below allow for significantly more latitude in the version
 * numbers than is allowed in the guidelines. No point in wasting user's
 * time enforcing them here. That's what flamewars are for.
 */
int
version_cmp(const char *pkg1, const char *pkg2)
{
	const char *c1, *c2, *v1, *v2;
	char *t1, *t2;
	int e1, e2, r1, r2, n1, n2;

	v1 = version_of(pkg1, &e1, &r1);
	v2 = version_of(pkg2, &e2, &r2);
	if (v1 == NULL || v2 == NULL)
		return (v1 == NULL ? (v2 == NULL ? 0 : -1) : 1);
	/* Minor optimisation. */
	if (strcmp(v1, v2) == 0)
		return 0;
	/* First compare epoch. */
	if (e1 != e2)
		return (e1 < e2 ? -1 : 1);
	else {
		/*
		 * We walk down the versions, trying to convert to numbers.
		 * We terminate when we reach an underscore, a comma or the
		 * string terminator, thanks to a nasty trick with strchr().
		 * strtol() conveniently gobbles up the chars it converts.
		 */
		c1 = strchr("_,", v1[0]);
		c2 = strchr("_,", v2[0]);
		while (c1 == NULL && c2 == NULL) {
			n1 = strtol(v1, &t1, 10);
			n2 = strtol(v2, &t2, 10);
			if (n1 != n2)
				return (n1 < n2 ? -1 : 1);
			/* The numbers are equal, check for letters. */
			c1 = strchr("_,.", t1[0]);
			c2 = strchr("_,.", t2[0]);
			if (c1 == NULL && c2 == NULL) {
				/* Both have letters. Compare them. */
				if (t1[0] != t2[0])
					return (t1[0] < t2[0] ? -1 : 1);
				/* Boring. The letters are equal. Carry on. */
				v1 = &t1[1], v2 = &t2[1];
			} else if (c1 == NULL) {
				/*
				 * Letters are strange. After a number, a letter counts
				 * as greater, but after a period it's less.
				 */
				return (isdigit(v1[0]) ? 1 : -1);
			} else if (c2 == NULL) {
				return (isdigit(v2[0]) ? -1 : 1);
			} else {
				/* Neither were letters.  Advance over the period. */
				v1 = (t1[0] == '.' ? &t1[1] : t1);
				v2 = (t2[0] == '.' ? &t2[1] : t2);
			}
			c1 = strchr("_,", v1[0]);
			c2 = strchr("_,", v2[0]);
		}
		/* If we got here, check if one version has something left. */
		if (c1 == NULL)
			return (isdigit(v1[0]) ? 1 : -1);
		if (c2 == NULL)
			return (isdigit(v2[0]) ? -1 : 1);
		/* We've run out of version. Try the revision... */
		if (r1 != r2)
			return (r1 < r2 ? -1 : 1);
		else
			return 0;
	}
}

/*
 * Rather obvious.
 */
char *
copy_string(const char *str)
{
	char *cp;

	if (str == NULL)
		return NULL;
	if ((cp = strdup(str)) == NULL) {
		msg_error(2, "%s: Unable to allocate memory", __func__);
		return NULL;
	}
	return cp;
}
