/*
 * XXX: (reg) insert final project name here.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Jeremy D. Lea
 * XXX: (reg) Insert date
 *
 * Based on pkg_install, by:
 *   Jordan K. Hubbard
 *   18 July 1993
 *
 * General routines for finding which package installed a file.
 */

#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include "list.h"
#include "lib.h"

/*
 * Allocate a new which list entry.
 */
which_entry *
new_which_entry(void)
{
	which_entry *w;

	if ((w = malloc(sizeof(which_entry))) == NULL) {
		msg_error(2, "%s: unable to allocate memory", __func__);
		return NULL;
	}
	bzero(w, sizeof(which_entry));
	return w;
}

/*
 * Free an entire which list.
 */
void
free_which(which_list *wl)
{
	which_entry *w;

	if (wl == NULL)
		return;
	while (!PLIST_EMPTY(wl)) {
		PLIST_REMOVE_HEAD(wl, w);
		free(w);
	}
}

/*
 * Add an item to a which list.
 */
which_entry *
add_which_entry(which_list *wl, enum which_t type, char *arg)
{
	char *cp;
	which_entry *w;
	
	if (wl == NULL || arg == NULL || strlen(arg) == 0)
		return NULL;
	if ((w = new_which_entry()) == NULL)
		return NULL;
	w->type = type;
	switch (w->type) {
	case WHICH_FILE:
		realpath(arg, w->file);
		break;

	case WHICH_EXEC:
    	/*
    	 * Use which(1) to find executables.  This only works if we don't
    	 * muck with the user's path...
    	 */
		if ((cp = vpipe("/usr/bin/which %s", arg)) != NULL) {
			realpath(cp, w->file);
			free(cp);
		}
		break;
		
	case WHICH_LIBRARY:
    	/*
    	 * Use ldconfig(1) to find libraries.  Taken from bsd.port.mk.
    	 */
		if ((cp = vpipe("ldconfig -r | grep -wF -e '-l%s' | sed -e 's/^.*=> //g'", arg)) != NULL) {
			realpath(cp, w->file);
			free(cp);
		}
		break;

	default:
		free(w);
		msg_error(2, "invalid which type in %s", __func__);
		return NULL;
	}
	/* Now check if we actually found a file... */
	if (strlen(w->file) == 0 || !isfile(w->file)) {
		free(w);
		return NULL;
	}
	PLIST_INSERT_TAIL(wl, w);
	return w;
}

/*
 * Add a whole pkg-plist to the which list.  This is done with the raw list,
 * rather than a processed one so that we can use it to scan for installed
 * files before a pkg_create(1).
 */
error_code
add_which_plist(which_list *wl, char *contents)
{
	FILE *fp;
	char *cp;
	char prefix[PATH_MAX], tmp[PATH_MAX], pline[PATH_MAX];
	plist_t type;

	if (wl == NULL || contents == NULL)
		return FAIL;
	snprintf(prefix, PATH_MAX, ".");
	/* Copy the packaging list, check it as we go */
	if (!strcmp(contents, "-"))
		fp = stdin;
	else
		fp = fopen(contents, "r");
	if (!fp) {
		msg_error(2, "unable to open contents file '%s' for input", contents);
		return FAIL;
	}
	while (read_line(pline, PATH_MAX, fp)) {
		type = read_plist_entry(pline, &cp);
		if (type == PLIST_CWD || type == PLIST_SRC)
			snprintf(prefix, PATH_MAX, "%s", cp);
		if (type == PLIST_FILE) {
			snprintf(tmp, PATH_MAX, "%s/%s", prefix, pline);
			add_which_entry(wl, WHICH_FILE, tmp);
		}
	}
	if (strcmp(contents, "-"))
		fclose(fp);
	return SUCCESS;
}

/*
 * Check our file against a package's plist, to see if we have a match.
 * Returns TRUE if we matched, FALSE otherwise.  We use this even though
 * we have the database, since we want to be sure.
 */
boolean
check_which(which_entry *w, package *pkg)
{
	plist_entry *p;
	char *dir = ".";
	char tmp[PATH_MAX], rp[PATH_MAX];

	if (w == NULL || pkg == NULL || strlen(w->file) == 0)
		return FALSE;
	w->pkg[0] = '\0';
	PLIST_FOREACH(p, &pkg->plist) {
		switch(p->type) {
		case PLIST_CWD:
			dir = p->name;
			break;
	
		case PLIST_FILE:
			if (p->ignore)
				break;
			snprintf(tmp, PATH_MAX, "%s/%s", dir, p->name);
			if (!isfile(tmp))
				break;
			realpath(tmp, rp);
			if (strcmp(rp, w->file) == 0) {
				strlcpy(w->pkg, pkg->name, PATH_MAX);
				return TRUE;
			}
			break;
	
		default:
			break;
		}
	}
	return FALSE;
}

/*
 * Scan an entry, looking from the matching package.  We try the database
 * first.  If it lies to us, we ignore it.  If we don't find the file, then
 * we take the expensive route and to a complete scan, on the assumption
 * that we are generally going to be looking for files which are managed, and
 * the database might be lying...
 */
boolean
scan_which_entry(which_entry *w)
{
	char tmp[PATH_MAX];
	pkg_entry *pe;

	if (w == NULL || strlen(w->file) == 0)
		return FALSE;
	if (!getcwd(tmp, PATH_MAX))
		snprintf(tmp, PATH_MAX, PkgMgr.wrkdir);
	if (which_filedb(w) != FAIL) {
		if (strlen(w->pkg) > 0) {
			enter_dir(tmp);
			return TRUE;
		}
	}
   	PLIST_FOREACH(pe, &PkgMgr) {
   		if (pe->status == PKG_REGISTERED) {
   			if (scan_pkg_entry(pe)) {
				enter_dir(tmp);
   				return FALSE;
			}
   			if (check_which(w, pe->pkg)) {
   				enter_dir(tmp);
				/* XXX: (reg) we should blow away the DB. */
   				return TRUE;
			}
			cleanup_pkg_entry(pe);
   		}
	}
	enter_dir(tmp);
	return FALSE;
}

/*
 * Scan all of the entries in a which_list, and return TRUE if we found at
 * least one.
 */
boolean
scan_which(which_list *wl)
{
	which_entry *w;
	boolean found = FALSE;

	if (wl == NULL)
		return FALSE;
	PLIST_FOREACH(w, wl) {
		if (scan_which_entry(w))
			found = TRUE;
	}
	return found;
}
